# -*- coding: utf-8 -*-

import uuid

import pytest

from py_gql.exc import InvalidValue, UnknownVariable
from py_gql.lang.parser import parse_value
from py_gql.schema.scalars import ID, UUID, Boolean, Float, Int, String
from py_gql.schema.types import (
    EnumType,
    EnumValue,
    InputField,
    InputObjectType,
    ListType,
    NonNullType,
)
from py_gql.utilities import untyped_value_from_ast, value_from_ast


class TestUntyped(object):
    @pytest.mark.parametrize(
        "value,expected",
        [
            ("null", None),
            ("true", True),
            ("false", False),
            ("123", 123),
            ("123.456", 123.456),
            ('"abc123"', "abc123"),
        ],
    )
    def test_it_parses_simple_values(self, value, expected):
        assert untyped_value_from_ast(parse_value(value)) == expected

    @pytest.mark.parametrize(
        "value,expected",
        [
            ("[true, false]", [True, False]),
            ("[true, 123.45]", [True, 123.45]),
            ("[true, null]", [True, None]),
            ('[true, ["foo", 1.2]]', [True, ["foo", 1.2]]),
        ],
    )
    def test_it_parses_list_values(self, value, expected):
        assert untyped_value_from_ast(parse_value(value)) == expected

    @pytest.mark.parametrize(
        "value,expected",
        [
            ("{ int: 123, bool: false }", {"int": 123, "bool": False}),
            ('{ foo: [ { bar: "baz"} ] }', {"foo": [{"bar": "baz"}]}),
        ],
    )
    def test_it_parses_input_objects(self, value, expected):
        assert untyped_value_from_ast(parse_value(value)) == expected

    @pytest.mark.parametrize(
        "value,expected",
        [
            ("TEST_ENUM_VALUE", "TEST_ENUM_VALUE"),
            ("[TEST_ENUM_VALUE]", ["TEST_ENUM_VALUE"]),
        ],
    )
    def test_it_parses_enum_values_as_plain_strings(self, value, expected):
        assert untyped_value_from_ast(parse_value(value)) == expected

    @pytest.mark.parametrize(
        "value,variables,expected",
        [
            ("$testVariable", {"testVariable": "foo"}, "foo"),
            ("[$testVariable]", {"testVariable": "foo"}, ["foo"]),
            ("{a:[$testVariable]}", {"testVariable": "foo"}, {"a": ["foo"]}),
            ("$testVariable", {"testVariable": None}, None),
        ],
    )
    def test_it_parses_variables(self, value, variables, expected):
        assert untyped_value_from_ast(parse_value(value), variables) == expected

    def test_it_raises_on_unknown_variables(self):
        with pytest.raises(UnknownVariable) as exc_info:
            untyped_value_from_ast(parse_value("$testVariable"), {})
        assert str(exc_info.value) == "testVariable"


Color = EnumType(
    "Color",
    [
        EnumValue("RED", 1, ""),
        EnumValue("GREEN", 2, ""),
        EnumValue("BLUE", 3, ""),
        EnumValue("NULL", None, ""),
    ],
)


ListOfBool = ListType(Boolean)
NonNullBool = NonNullType(Boolean)
ListOfNonNullBool = ListType(NonNullType(Boolean))
NonNullListOfBool = NonNullType(ListType(Boolean))
NonNullListOfNonNullBool = NonNullType(ListType(NonNullType(Boolean)))


TestInput = InputObjectType(
    "TestInput",
    [
        InputField("int", Int, default_value=42),
        InputField("bool", Boolean),
        InputField("requiredBool", NonNullBool),
    ],
)


class TestTyped(object):
    def _run_test_case(self, value, type_, expected, error, variables=None):
        if error is None:
            assert (
                value_from_ast(parse_value(value), type_, variables) == expected
            )
        else:
            with pytest.raises(error):
                value_from_ast(parse_value(value), type_, variables)

    @pytest.mark.parametrize(
        "type_,value,expected",
        [
            (Boolean, "true", True),
            (Boolean, "false", False),
            (Int, "123", 123),
            (Float, "123", 123),
            (Float, "123.456", 123.456),
            (String, '"abc123"', "abc123"),
            (ID, "123456", "123456"),
            (ID, '"123456"', "123456"),
            (
                UUID,
                '"aacefca9-41c6-4a01-b372-632b4dc2506c"',
                uuid.UUID("aacefca9-41c6-4a01-b372-632b4dc2506c"),
            ),
            (
                UUID,
                '"AACEFCA9-41C6-4A01-B372-632B4DC2506C"',
                uuid.UUID("aacefca9-41c6-4a01-b372-632b4dc2506c"),
            ),
        ],
    )
    def test_it_converts_according_to_input_coercion_rules(
        self, type_, value, expected
    ):
        self._run_test_case(value, type_, expected, None)

    @pytest.mark.parametrize(
        "type_,value",
        [
            (Boolean, "123"),
            (Int, "123.456"),
            (Int, "true"),
            (Int, '"123"'),
            (Float, '"123"'),
            (String, "123"),
            (String, "true"),
            (ID, "123.456"),
            (UUID, "123.456"),
        ],
    )
    def test_it_does_not_convert_when_input_coercion_rules_reject_a_value(
        self, type_, value
    ):
        self._run_test_case(value, type_, None, InvalidValue)

    @pytest.mark.parametrize(
        "value,expected,error",
        [
            ("RED", 1, None),
            ("BLUE", 3, None),
            ("3", None, InvalidValue),
            ('"BLUE"', None, InvalidValue),
            ("NULL", None, None),
            ("UNDEFINED", None, InvalidValue),
        ],
    )
    def test_it_converts_enum_values_according_to_input_coercion_rules(
        self, value, expected, error
    ):
        self._run_test_case(value, Color, expected, error)

    def test_it_coerces_nullable_to_none(self):
        assert value_from_ast(parse_value("null"), Boolean, None) is None

    def test_it_raises_on_non_nullable_null(self):
        with pytest.raises(InvalidValue):
            value_from_ast(parse_value("null"), NonNullType(Boolean), None)

    @pytest.mark.parametrize(
        "type_,value,expected,error",
        [
            (ListOfBool, "true", [True], None),
            (ListOfBool, "[true]", [True], None),
            (ListOfBool, "123", None, InvalidValue),
            (ListOfBool, "null", None, None),
            (ListOfBool, "[true, false]", [True, False], None),
            (ListOfBool, "[true, 123]", None, InvalidValue),
            (ListOfBool, "[true, null]", [True, None], None),
            (ListOfBool, "{ true: true }", None, InvalidValue),
        ],
    )
    def test_it_coerces_lists_of_values(self, type_, value, expected, error):
        self._run_test_case(value, type_, expected, error)

    @pytest.mark.parametrize(
        "type_,value,expected,error",
        [
            (ListOfBool, "true", [True], None),
            (ListOfBool, "[true]", [True], None),
            (NonNullListOfBool, "123", None, InvalidValue),
            (NonNullListOfBool, "null", None, InvalidValue),
            (NonNullListOfBool, "[true, false]", [True, False], None),
            (NonNullListOfBool, "[true, 123]", None, InvalidValue),
            (NonNullListOfBool, "[true, null]", [True, None], None),
        ],
    )
    def test_it_coerces_non_null_lists_of_values(
        self, type_, value, expected, error
    ):
        self._run_test_case(value, type_, expected, error)

    @pytest.mark.parametrize(
        "type_,value,expected,error",
        [
            (ListOfBool, "true", [True], None),
            (ListOfBool, "[true]", [True], None),
            (ListOfNonNullBool, "123", None, InvalidValue),
            (ListOfNonNullBool, "null", None, None),
            (ListOfNonNullBool, "[true, false]", [True, False], None),
            (ListOfNonNullBool, "[true, 123]", None, InvalidValue),
            (ListOfNonNullBool, "[true, null]", None, InvalidValue),
        ],
    )
    def test_it_coerces_lists_of_non_null_values(
        self, type_, value, expected, error
    ):
        self._run_test_case(value, type_, expected, error)

    @pytest.mark.parametrize(
        "type_,value,expected,error",
        [
            (ListOfBool, "true", [True], None),
            (ListOfBool, "[true]", [True], None),
            (NonNullListOfNonNullBool, "123", None, InvalidValue),
            (NonNullListOfNonNullBool, "null", None, InvalidValue),
            (NonNullListOfNonNullBool, "[true, false]", [True, False], None),
            (NonNullListOfNonNullBool, "[true, 123]", None, InvalidValue),
            (NonNullListOfNonNullBool, "[true, null]", None, InvalidValue),
        ],
    )
    def test_it_coerces_non_null_lists_of_non_null_values(
        self, type_, value, expected, error
    ):
        self._run_test_case(value, type_, expected, error)

    @pytest.mark.parametrize(
        "type_,value,expected,error",
        [
            (TestInput, "null", None, None),
            (TestInput, "123", None, InvalidValue),
            (TestInput, "[]", None, InvalidValue),
            (
                TestInput,
                "{ int: 123, requiredBool: false }",
                {"int": 123, "requiredBool": False},
                None,
            ),
            (
                TestInput,
                "{ bool: true, requiredBool: false }",
                {"int": 42, "bool": True, "requiredBool": False},
                None,
            ),
            (
                TestInput,
                "{ int: true, requiredBool: true }",
                None,
                InvalidValue,
            ),
            (TestInput, "{ requiredBool: null }", None, InvalidValue),
            (TestInput, "{ bool: true }", None, InvalidValue),
        ],
    )
    def test_it_coerces_input_object(self, type_, value, expected, error):
        self._run_test_case(value, type_, expected, error)

    @pytest.mark.parametrize(
        "type_,value,variables,expected,error",
        [
            (Boolean, "$var", {}, None, InvalidValue),
            (Boolean, "$var", {"var": True}, True, None),
            (Boolean, "$var", {"var": None}, None, None),
        ],
    )
    def test_it_accepts_variable_values_assuming_already_coerced(
        self, type_, value, variables, expected, error
    ):
        self._run_test_case(value, type_, expected, error, variables)

    def test_it_raises_on_null_variable_with_non_null_type(self):
        with pytest.raises(InvalidValue) as exc_info:
            value_from_ast(
                parse_value("$testVariable"),
                NonNullBool,
                {"testVariable": None},
            )

        assert str(exc_info.value) == (
            'Variable "$testVariable" used for type "Boolean!" '
            "must not be null."
        )

    @pytest.mark.parametrize(
        "type_,value,variables,expected,error",
        [
            # We fail hard on missing variables
            # (ListOfBool, ' [ $foo ]', {}, [None], None),
            (ListOfBool, " [ $foo ]", {}, None, UnknownVariable),
            (ListOfNonNullBool, " [ $foo ]", {}, None, InvalidValue),
            (ListOfNonNullBool, " [ $foo ]", {"foo": True}, [True], None),
            # Note: variables are expected to have already been coerced, so we
            # do not expect the singleton wrapping behavior for variables.
            (ListOfNonNullBool, "$foo", {"foo": True}, True, None),
            (ListOfNonNullBool, "$foo", {"foo": [True]}, [True], None),
        ],
    )
    def test_it_asserts_variables_are_provided_as_items_in_lists(
        self, type_, value, variables, expected, error
    ):
        self._run_test_case(value, type_, expected, error, variables)

    @pytest.mark.parametrize(
        "type_,value,variables,expected,error",
        [
            # We fail hard on missing variables
            # (TestInput, '{ int: $foo, bool: $foo, requiredBool: true }', {}, {
            #     'int': 42,
            #     'requiredBool': True,
            # }, None),
            (
                TestInput,
                "{ int: $foo, bool: $foo, requiredBool: true }",
                {},
                None,
                UnknownVariable,
            ),
            (TestInput, "{ requiredBool: $foo }", {}, None, UnknownVariable),
            (
                TestInput,
                "{ requiredBool: $foo, bool: $foo }",
                {"foo": True},
                {"int": 42, "requiredBool": True, "bool": True},
                None,
            ),
        ],
    )
    def test_it_omits_input_object_fields_for_unprovided_variables(
        self, type_, value, variables, expected, error
    ):
        self._run_test_case(value, type_, expected, error, variables)



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

from py_gql.validation.rules import KnownTypeNamesChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_known_type_names_are_valid(schema):
    run_test(
        KnownTypeNamesChecker,
        schema,
        """
        query Foo($var: String, $required: [String!]!) {
            user(id: 4) {
                pets { ... on Pet { name }, ...PetFields, ... { name } }
            }
        }
        fragment PetFields on Pet {
            name
        }
        """,
    )


def test_unknown_type_names_are_invalid(schema):
    run_test(
        KnownTypeNamesChecker,
        schema,
        """
        query Foo($var: JumbledUpLetters) {
            user(id: 4) {
                name
                pets { ... on Badger { name }, ...PetFields }
            }
        }
        fragment PetFields on Peettt {
            name
        }
        """,
        [
            'Unknown type "JumbledUpLetters"',
            # 'Unknown type "Badger"',
            # 'Unknown type "Peettt"',
        ],
    )


def test_ignores_type_definitions(schema):
    run_test(
        KnownTypeNamesChecker,
        schema,
        """
        type NotInTheSchema {
            field: FooBar
        }
        interface FooBar {
            field: NotInTheSchema
        }
        union U = A | B
        input Blob {
            field: UnknownType
        }
        query Foo($var: NotInTheSchema) {
            user(id: $var) {
            id
            }
        }
        """,
        ['Unknown type "NotInTheSchema"'],
    )



# -*- coding: utf-8 -*-


from py_gql._utils import find_one
from py_gql.exc import UnknownEnumValue, UnknownType
from py_gql.lang.visitor import DispatchingVisitor
from py_gql.schema import (
    EnumType,
    InputObjectType,
    InterfaceType,
    ListType,
    ObjectType,
    is_composite_type,
    is_input_type,
    is_output_type,
    nullable_type,
    unwrap_type,
)
from py_gql.schema.introspection import (
    schema_field,
    type_field,
    type_name_field,
)


def _peek(lst, count=1, default=None):
    return lst[-1 * count] if len(lst) >= count else default


def _or_none(value, predicate=bool):
    return value if predicate(value) else None


def _get_field_def(schema, parent_type, field):
    name = field.name.value
    if parent_type is schema.query_type:
        if name == schema_field.name:
            return schema_field
        if name == type_field.name:
            return type_field

    if is_composite_type(parent_type) and name == type_name_field.name:
        return type_name_field

    if isinstance(parent_type, (ObjectType, InterfaceType)):
        return parent_type.field_map.get(name, None)

    return None


# This is a very basic re-implementation of the reference javascript
# implementation which is compatible with our version of AST visitors
# and it can most likley be improved.
class TypeInfoVisitor(DispatchingVisitor):
    """ Visitor that tracks current types while traversing a Document.

    All tracked types are considered with regards to the provided schema,
    however unknown types and other unexpected errors will be downgraded to
    null values in order to not crash the traversal. This leaves the consumer
    responsible to handle such cases.

    Warning:
        When using this alongside other visitors (such as when using
        :class:`py_gql.lang.visitor.ParrallelVisitor`), this visitor **needs**
        to be the first one to visit the nodes in order for the information
        provided donwstream to be accurate.
    """

    __slots__ = (
        "_schema",
        "_type_stack",
        "_input_type_stack",
        "_parent_type_stack",
        "_field_stack",
        "_input_value_def_stack",
        "directive",
        "argument",
        "enum_value",
    )

    # pylint: disable = super-init-not-called
    def __init__(self, schema, _get_field_def=None):
        self._schema = schema

        self._type_stack = []
        self._parent_type_stack = []
        self._input_type_stack = []
        self._field_stack = []
        self._input_value_def_stack = []

        #: Optional[:class:`py_gql.schema.Directive`]: Current directive if applicable
        self.directive = None
        #: Optional[:class:`py_gql.schema.Argument`]: Current argument if applicable
        self.argument = None
        #: Optional[:class:`py_gql.schema.EnumValue`]: Current enum value if applicable
        self.enum_value = None

    @property
    def type(self):
        """ Optional[:class:`py_gql.schema.Type`]: Current type if applicable
        """
        return _peek(self._type_stack)

    @property
    def parent_type(self):
        """ Optional[:class:`py_gql.schema.Type`]: Current type if applicable
        """
        return _peek(self._parent_type_stack, 1)

    @property
    def input_type(self):
        """ Optional[:class:`py_gql.schema.Type`]: Current input type if
        applicable (when visiting arguments) """
        return _peek(self._input_type_stack, 1)

    @property
    def parent_input_type(self):
        """ Optional[:class:`py_gql.schema.Type`]: Current parent input type if
        applicable (when visiting input objects) """
        return _peek(self._input_type_stack, 2)

    @property
    def field(self):
        """ Optional[:class:`py_gql.schema.Field`]: Current field definition if
        applicable (when visiting object) """
        return _peek(self._field_stack)

    @property
    def input_value_def(self):
        """ Optional[Union[:class:`py_gql.schema.Argument`,
        :class:`py_gql.schema.InputField`]]: Current input value definition
        (e.g. arg def, input field) if applicable """
        return _peek(self._input_value_def_stack)

    def _get_field_def(self, node):
        parent_type = self.parent_type
        return (
            _get_field_def(self._schema, parent_type, node)
            if parent_type
            else None
        )

    def _type_from_ast(self, type_node):
        try:
            return self._schema.get_type_from_literal(type_node)
        except UnknownType:
            return None

    def _leave_input_value(self):
        self._input_type_stack.pop()
        self._input_value_def_stack.pop()

    def enter_selection_set(self, node):
        named_type = unwrap_type(self.type)
        self._parent_type_stack.append(_or_none(named_type, is_composite_type))

    def leave_selection_set(self, node):
        self._parent_type_stack.pop()

    def enter_field(self, node):
        field_def = self._get_field_def(node)
        self._field_stack.append(field_def)
        if field_def:
            self._type_stack.append(_or_none(field_def.type, is_output_type))
        else:
            self._type_stack.append(None)

    def leave_field(self, node):
        self._type_stack.pop()
        self._field_stack.pop()

    def enter_directive(self, node):
        self.directive = self._schema.directives.get(node.name.value)

    def leave_directive(self, node):
        self.directive = None

    def enter_operation_definition(self, node):
        type_ = {
            "query": self._schema.query_type,
            "mutation": self._schema.mutation_type,
            "subscription": self._schema.subscription_type,
        }.get(node.operation, None)
        self._type_stack.append(
            type_ if isinstance(type_, ObjectType) else None
        )

    def leave_operation_definition(self, node):
        self._type_stack.pop()

    def enter_fragment_definition(self, node):
        self._type_stack.append(
            _or_none(self._type_from_ast(node.type_condition), is_output_type)
        )

    def leave_fragment_definition(self, node):
        self._type_stack.pop()

    def enter_inline_fragment(self, node):
        if node.type_condition:
            self._type_stack.append(
                _or_none(
                    self._type_from_ast(node.type_condition), is_output_type
                )
            )
        else:
            self._type_stack.append(_or_none(self.type, is_output_type))

    def leave_inline_fragment(self, node):
        self._type_stack.pop()

    def enter_variable_definition(self, node):
        self._input_type_stack.append(
            _or_none(self._type_from_ast(node.type), is_input_type)
        )

    def leave_variable_definition(self, node):
        self._input_type_stack.pop()

    def enter_argument(self, node):
        ctx = self.directive or self.field
        if ctx:
            name = node.name.value
            self.argument = find_one(ctx.args, lambda a: a.name == name)
            self._input_value_def_stack.append(self.argument)
            self._input_type_stack.append(
                self.argument.type
                if self.argument and is_input_type(self.argument.type)
                else None
            )
        else:
            self.argument = None
            self._input_type_stack.append(None)
            self._input_value_def_stack.append(None)

    def leave_argument(self, node):
        self.argument = None
        self._leave_input_value()

    def enter_list_value(self, node):
        list_type = nullable_type(self.input_type)
        item_type = (
            unwrap_type(list_type) if isinstance(list_type, ListType) else None
        )
        self._input_type_stack.append(_or_none(item_type, is_input_type))
        # List positions never have a default value.
        self._input_value_def_stack.append(None)

    def leave_list_value(self, node):
        self._leave_input_value()

    def enter_object_field(self, node):
        object_type = unwrap_type(self.input_type)
        if isinstance(object_type, InputObjectType):
            name = node.name.value
            field_def = find_one(object_type.fields, lambda f: f.name == name)
            self._input_value_def_stack.append(field_def)
            self._input_type_stack.append(
                field_def.type
                if field_def and is_input_type(field_def.type)
                else None
            )
        else:
            self._input_type_stack.append(None)
            self._input_value_def_stack.append(None)

    def leave_object_field(self, node):
        self._leave_input_value()

    def enter_enum_value(self, node):
        enum = unwrap_type(self.input_type)
        if isinstance(enum, EnumType):
            try:
                self.enum_value = enum.get_value(node.value)
            except UnknownEnumValue:
                self.enum_value = None

    def leave_enum_value(self, node):
        self.enum_value = None



# -*- coding: utf-8 -*-
""" Default directives. """
from __future__ import unicode_literals

from .scalars import Boolean, String
from .types import Argument, Directive, NonNullType

IncludeDirective = Directive(
    "include",
    description=(
        "Directs the executor to include this field or fragment only when "
        "the `if` argument is true."
    ),
    locations=["FIELD", "FRAGMENT_SPREAD", "INLINE_FRAGMENT"],
    args=[
        Argument("if", NonNullType(Boolean), description="Included when true")
    ],
)

SkipDirective = Directive(
    "skip",
    description=(
        "Directs the executor to skip this field or fragment when the `if` "
        "argument is true."
    ),
    locations=["FIELD", "FRAGMENT_SPREAD", "INLINE_FRAGMENT"],
    args=[
        Argument("if", NonNullType(Boolean), description="Skipped when true")
    ],
)

DEFAULT_DEPRECATION = "No longer supported"

DeprecatedDirective = Directive(
    "deprecated",
    description="Marks an element of a GraphQL schema as no longer supported.",
    locations=["FIELD_DEFINITION", "ENUM_VALUE"],
    args=[
        Argument(
            "reason",
            String,
            default_value=DEFAULT_DEPRECATION,
            description=(
                "Explains why this element was deprecated, usually also "
                "including a suggestion for how to access supported "
                "similar data. Formatted in [Markdown](https://daringfireball"
                ".net/projects/markdown/)."
            ),
        )
    ],
)


# These are the types which are part of the spec and will always be available
# in any spec compliant GraphQL server.
SPECIFIED_DIRECTIVES = (IncludeDirective, SkipDirective, DeprecatedDirective)



# -*- coding: utf-8 -*-
""" """

from .exc import ExecutionError, GraphQLSyntaxError, VariablesCoercionError
from .execution import GraphQLResult, GraphQLTracer, _concurrency, execute
from .execution.tracing import NullTracer
from .lang import parse
from .schema import Schema
from .validation import SPECIFIED_RULES, validate_ast


def _graphql(
    schema,
    document,
    variables=None,
    operation_name=None,
    initial_value=None,
    validators=None,
    context=None,
    middlewares=None,
    tracer=None,
    executor=None,
):
    """ Main GraphQL entrypoint encapsulating query processing from start to
    finish.

    Args:
        schema (py_gql.schema.Schema): Schema to execute the query against

        document (str): The query document

        variables (Optional[dict]):
            Raw, JSON decoded variables parsed from the request

        operation_name (Optional[str]): Operation to execute
            If specified, the operation with the given name will be executed.
            If not, this executes the single operation without disambiguation.

        initial_value (Any): Root resolution value passed to all top-level
            resolvers

        validators (Optional[List[py_gql.validation.ValidationVisitor]]):
            Custom validators.
            Setting this will replace the defaults so if you just want to add
            some rules, append to :obj:`py_gql.validation.SPECIFIED_RULES`.

        context (Any): Custom application-specific execution context
            Use this to pass in anything your resolvers require like database
            connection, user information, etc.
            Limits on the type(s) used here will depend on your own resolver
            implementations and the executor class you use. Most thread safe
            data-structures should work.

        middlewares (Optional[List[Callable]]):
            List of middleware callable to use when resolving fields

        tracer (Optional[py_gql.GraphQLTracer]): Tracer instance

        executor (Optional[py_gql.execution.Executor]): Executor instance

    Returns:
        Future: Deferred execution result
    """

    assert isinstance(schema, Schema)
    schema.validate()

    tracer = tracer or NullTracer()
    assert isinstance(tracer, GraphQLTracer)

    if not isinstance(tracer, NullTracer):
        middlewares = [tracer._middleware] + (middlewares or [])

    try:
        tracer.trace(
            "query",
            "start",
            document=document,
            variables=variables,
            operation_name=operation_name,
        )

        with tracer._trace_context("parse", document=document):
            ast = parse(document, allow_type_system=False)

        validators = SPECIFIED_RULES if validators is None else validators

        with tracer._trace_context("validate", ast=ast):
            validation_result = validate_ast(schema, ast, validators=validators)

        if not validation_result:
            tracer.trace(
                "query",
                "end",
                document=document,
                variables=variables,
                operation_name=operation_name,
            )
            return _concurrency.deferred(
                GraphQLResult(errors=validation_result.errors),
                cls=_concurrency.DummyFuture,
            )
        else:
            tracer.trace("execute", "start", ast=ast, variables=variables)
            result = execute(
                schema,
                ast,
                executor=executor,
                initial_value=initial_value,
                context_value=context,
                variables=variables,
                operation_name=operation_name,
                middlewares=middlewares,
            )

            def close_trace(_):
                tracer.trace("execute", "end", ast=ast, variables=variables)
                tracer.trace(
                    "query",
                    "end",
                    document=document,
                    variables=variables,
                    operation_name=operation_name,
                )

            result.add_done_callback(close_trace)
            return result

    except GraphQLSyntaxError as err:
        tracer.trace(
            "query",
            "end",
            document=document,
            variables=variables,
            operation_name=operation_name,
        )
        return _concurrency.deferred(
            GraphQLResult(errors=[err]), cls=_concurrency.DummyFuture
        )
    except VariablesCoercionError as err:
        tracer.trace(
            "query",
            "end",
            document=document,
            variables=variables,
            operation_name=operation_name,
        )
        return _concurrency.deferred(
            GraphQLResult(data=None, errors=err.errors),
            cls=_concurrency.DummyFuture,
        )
    except ExecutionError as err:
        tracer.trace(
            "query",
            "end",
            document=document,
            variables=variables,
            operation_name=operation_name,
        )
        return _concurrency.deferred(
            GraphQLResult(data=None, errors=[err]), cls=_concurrency.DummyFuture
        )


def graphql(
    schema,
    document,
    variables=None,
    operation_name=None,
    initial_value=None,
    validators=None,
    context=None,
    middlewares=None,
    tracer=None,
    executor=None,
    timeout=None,
):
    """ This is the main entrypoint for execution of GraphQL queries.

    Args:
        schema (py_gql.schema.Schema): Schema to execute the query against

        document (str): The query document

        variables (Optional[dict]):
            Raw, JSON decoded variables parsed from the request

        operation_name (Optional[str]): Operation to execute
            If specified, the operation with the given name will be executed.
            If not, this executes the single operation without disambiguation.

        initial_value (Any): Root resolution value passed to all top-level
            resolvers

        validators (Optional[List[py_gql.validation.ValidationVisitor]]):
            Custom validators.
            Setting this will replace the defaults so if you just want to add
            some rules, append to :obj:`py_gql.validation.SPECIFIED_RULES`.

        context (Any): Custom application-specific execution context
            Use this to pass in anything your resolvers require like database
            connection, user information, etc.
            Limits on the type(s) used here will depend on your own resolver
            implementations and the executor class you use. Most thread safe
            data-structures should work.

        middlewares (Optional[List[Callable]]):
            List of middleware callable to use when resolving fields

        tracer (Optional[py_gql.GraphQLTracer]): Tracer instance

        executor (Optional[py_gql.execution.Executor]): Executor instance

        timeout (Union[float,int]): blocking timeout in seconds
            When using custom executors which support non-blocking / parallel
            execution, the call will block until the execution result is
            available or ``timeout`` is exceeded, however by default timeout
            has no effect as the resolution is blocking.

    Returns:
        py_gql.GraphQLResult: Execution result
    """
    result = _graphql(
        schema,
        document,
        variables=variables,
        operation_name=operation_name,
        initial_value=initial_value,
        validators=validators,
        context=context,
        middlewares=middlewares,
        tracer=tracer,
        executor=executor,
    )

    return result.result(timeout=timeout)



# -*- coding: utf-8 -*-
""" Tests for extend schema from a pre-built schema. """
# This file only test some cases that are not exercised when testing
# make_executable_schema as well as adds some lower level assertions.

import pytest

from py_gql._string_utils import dedent
from py_gql.exc import SDLError
from py_gql.lang import ast as _ast
from py_gql.schema import (
    Argument,
    Directive,
    Field,
    InputField,
    InputObjectType,
    Int,
    ObjectType,
    Schema,
)
from py_gql.schema.build import extend_schema

BASE_SCHEMA = Schema(ObjectType("Query", [Field("foo", Int)]))


def test_noop_without_extension_nodes():
    # NOTE: The parser cannot technically build such a document
    new_schema = extend_schema(BASE_SCHEMA, _ast.Document(definitions=[]))
    assert new_schema is BASE_SCHEMA


def test_raises_on_schema_definition_in_strict_mode():
    with pytest.raises(SDLError):
        extend_schema(
            BASE_SCHEMA,
            """
            schema {
                query: Query
            }
            """,
        )


def test_raises_on_known_type_in_strict_mode():
    with pytest.raises(SDLError):
        extend_schema(BASE_SCHEMA, "scalar String")


def test_ignores_known_type_in_non_strict_mode():
    new_schema = extend_schema(BASE_SCHEMA, "scalar String", strict=False)
    assert new_schema is BASE_SCHEMA


def test_raises_on_known_directive_in_strict_mode():
    with pytest.raises(SDLError):
        extend_schema(
            BASE_SCHEMA,
            "directive @skip(if: Boolean!)"
            "on FIELD | FRAGMENT_SPREAD | INLINE_FRAGMENT",
        )


def test_ignores_known_directive_in_non_strict_mode():
    new_schema = extend_schema(
        BASE_SCHEMA,
        "directive @skip(if: Boolean!)"
        "on FIELD | FRAGMENT_SPREAD | INLINE_FRAGMENT",
        strict=False,
    )
    assert new_schema is BASE_SCHEMA


def test_raises_on_unknown_type_in_strict_mode():
    with pytest.raises(SDLError):
        extend_schema(BASE_SCHEMA, "extend scalar UUID @foo")


def test_ignores_unknown_type_in_non_strict_mode():
    new_schema = extend_schema(
        BASE_SCHEMA, "extend scalar UUID @foo", strict=False
    )
    assert new_schema is BASE_SCHEMA


def test_it_adds_new_type_definitions_and_opeations_to_schema():
    assert (
        extend_schema(
            BASE_SCHEMA,
            """
            scalar UUID

            type SomeMutation {
                bar: UUID
            }

            extend schema {
                mutation: SomeMutation
            }
            """,
        ).to_string()
        == dedent(
            """
            schema {
                query: Query
                mutation: SomeMutation
            }

            type Query {
                foo: Int
            }

            type SomeMutation {
                bar: UUID
            }

            scalar UUID
            """
        )
    )


def test_it_rejects_duplicate_operation():
    with pytest.raises(SDLError):
        extend_schema(
            BASE_SCHEMA,
            """
            type NewQuery { bar: String }
            extend schema { query: NewQuery }
            """,
        )


def test_it_adds_new_directives_to_schema():
    assert extend_schema(
        BASE_SCHEMA, "directive @some(a: Boolean!) on FIELD"
    ).to_string() == dedent(
        """
        directive @some(a: Boolean!) on FIELD

        type Query {
            foo: Int
        }
        """
    )


def test_it_correctly_updates_references():
    arg_type = InputObjectType("Bar", [InputField("a", Int)])
    schema_with_args = Schema(
        query_type=ObjectType(
            "Query",
            [
                Field(
                    "foo",
                    ObjectType("Foo", [Field("a", Int)]),
                    [Argument("bar", arg_type)],
                )
            ],
        ),
        directives=[Directive("baz", ["FIELD"], [Argument("bar", arg_type)])],
    )

    update_schema = extend_schema(
        schema_with_args,
        """
        extend input Bar {
            b: String
        }

        extend type Foo {
            b: String
        }
        """,
    )

    assert update_schema.to_string() == dedent(
        """
        directive @baz(bar: Bar) on FIELD

        input Bar {
            a: Int
            b: String
        }

        type Foo {
            a: Int
            b: String
        }

        type Query {
            foo(bar: Bar): Foo
        }
        """
    )

    field_type = update_schema.get_type("Query").field_map["foo"].type
    root_field_type = update_schema.get_type("Foo")

    assert root_field_type is field_type

    field_arg_type = (
        update_schema.get_type("Query").field_map["foo"].arg_map["bar"].type
    )
    directive_arg_type = update_schema.directives.get("baz").arg_map["bar"].type
    root_arg_type = update_schema.get_type("Bar")

    assert field_arg_type is root_arg_type
    assert directive_arg_type is root_arg_type



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import SingleFieldSubscriptionsChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_valid_subscription(schema):
    run_test(
        SingleFieldSubscriptionsChecker,
        schema,
        """
    subscription ImportantEmails {
        importantEmails
    }""",
    )


def test_fails_with_more_than_one_root_field(schema):
    run_test(
        SingleFieldSubscriptionsChecker,
        schema,
        """
    subscription ImportantEmails {
        importantEmails
        notImportantEmails
    }""",
        [
            'Subscription "ImportantEmails" must select only one '
            "top level field."
        ],
    )


def test_fails_with_more_than_one_root_field_including_introspection(schema):
    run_test(
        SingleFieldSubscriptionsChecker,
        schema,
        """
    subscription ImportantEmails {
        importantEmails
        __typename
    }
    """,
        [
            'Subscription "ImportantEmails" must select only one '
            "top level field."
        ],
    )


def test_fails_with_many_more_than_one_root_field(schema):
    run_test(
        SingleFieldSubscriptionsChecker,
        schema,
        """
    subscription ImportantEmails {
        importantEmails
        notImportantEmails
        spamEmails
        }
    """,
        [
            'Subscription "ImportantEmails" must select only one '
            "top level field."
        ],
    )


def test_fails_with_more_than_one_root_field_in_anonymous_subscriptions(schema):
    run_test(
        SingleFieldSubscriptionsChecker,
        schema,
        """
    subscription {
        importantEmails
        notImportantEmails
    }
    """,
        ['Subscription "" must select only one ' "top level field."],
    )



# -*- coding: utf-8 -*-

import pytest

from py_gql.lang import ast as _ast
from py_gql.schema import (
    ID,
    Boolean,
    EnumType,
    Float,
    InputField,
    InputObjectType,
    Int,
    ListType,
    NonNullType,
    ScalarType,
    String,
)
from py_gql.schema.scalars import MAX_INT
from py_gql.utilities import ast_node_from_value


def _custom_serialize(x):
    if x == 42:
        return None
    return x


CustomScalar = ScalarType(
    name="CustomScalar", serialize=_custom_serialize, parse=lambda x: x
)


@pytest.mark.parametrize(
    "value, input_type, expected",
    [
        (True, Boolean, _ast.BooleanValue(value=True)),
        (False, Boolean, _ast.BooleanValue(value=False)),
        (None, Boolean, _ast.NullValue()),
        (1, Boolean, _ast.BooleanValue(value=True)),
        (0, Boolean, _ast.BooleanValue(value=False)),
        (0, NonNullType(Boolean), _ast.BooleanValue(value=False)),
        (
            None,
            NonNullType(Boolean),
            ValueError('Value of type "Boolean!" cannot be null'),
        ),
        (-1, Int, _ast.IntValue(value="-1")),
        (123.0, Int, _ast.IntValue(value="123")),
        (1e4, Int, _ast.IntValue(value="10000")),
        (
            123.5,
            Int,
            ValueError("Int cannot represent non-integer value: 123.5"),
        ),
        (
            1e40,
            Int,
            ValueError("Int cannot represent non 32-bit signed integer: 1e+40"),
        ),
        (-1, Float, _ast.IntValue(value="-1")),
        (123.0, Float, _ast.IntValue(value="123")),
        (123.5, Float, _ast.FloatValue(value="123.5")),
        (1e4, Float, _ast.IntValue(value="10000")),
        (1e40, Float, _ast.FloatValue(value="1e+40")),
        ("hello", String, _ast.StringValue(value="hello")),
        ("VALUE", String, _ast.StringValue(value="VALUE")),
        ("VA\nLUE", String, _ast.StringValue(value="VA\nLUE")),
        (123, String, _ast.StringValue(value="123")),
        (False, String, _ast.StringValue(value="false")),
        (None, String, _ast.NullValue()),
        ("hello", ID, _ast.StringValue(value="hello")),
        (-1, ID, _ast.IntValue(value="-1")),
        (123, ID, _ast.IntValue(value="123")),
        ("01", ID, _ast.StringValue(value="01")),
        (None, ID, _ast.NullValue()),
        (42, CustomScalar, _ast.NullValue()),
        ("42.42", CustomScalar, _ast.FloatValue(value="42.42")),
        (42.42, CustomScalar, _ast.FloatValue(value="42.42")),
        (MAX_INT + 2, CustomScalar, _ast.FloatValue(value="2147483649")),
        ("foo", CustomScalar, _ast.StringValue(value="foo")),
    ],
)
def test_ast_node_from_value_with_scalars(value, input_type, expected):
    if isinstance(expected, Exception):
        with pytest.raises(type(expected)) as exc_info:
            ast_node_from_value(value, input_type)
        assert str(expected) == str(exc_info.value)
    else:
        assert ast_node_from_value(value, input_type) == expected


# Must still be hashable to work with enums
complex_value = ("someArbitrary", "complexValue")

Enum = EnumType("MyEnum", ["HELLO", "GOODBYE", ("COMPLEX", complex_value)])


@pytest.mark.parametrize(
    "value, expected",
    [
        ("HELLO", _ast.EnumValue(value="HELLO")),
        ("GOODBYE", _ast.EnumValue(value="GOODBYE")),
        (complex_value, _ast.EnumValue(value="COMPLEX")),
        ("hello", ValueError("Invalid value 'hello' for enum MyEnum")),
        ("VALUE", ValueError("Invalid value 'VALUE' for enum MyEnum")),
    ],
)
def test_ast_node_from_value_with_enums(value, expected):
    if isinstance(expected, Exception):
        with pytest.raises(type(expected)) as exc_info:
            ast_node_from_value(value, Enum)
        assert str(expected) == str(exc_info.value)
    else:
        assert ast_node_from_value(value, Enum) == expected


@pytest.mark.parametrize(
    "value, input_type, expected",
    [
        (
            ["FOO", "BAR"],
            ListType(String),
            _ast.ListValue(
                values=[
                    _ast.StringValue(value="FOO"),
                    _ast.StringValue(value="BAR"),
                ]
            ),
        ),
        (
            ["HELLO", "GOODBYE"],
            ListType(Enum),
            _ast.ListValue(
                values=[
                    _ast.EnumValue(value="HELLO"),
                    _ast.EnumValue(value="GOODBYE"),
                ]
            ),
        ),
        ("FOO", ListType(String), _ast.StringValue(value="FOO")),
    ],
)
def test_ast_node_from_value_with_list_types(value, input_type, expected):
    if isinstance(expected, Exception):
        with pytest.raises(type(expected)) as exc_info:
            ast_node_from_value(value, input_type)
        assert str(expected) == str(exc_info.value)
    else:
        assert ast_node_from_value(value, input_type) == expected


InputObject = InputObjectType(
    "MyInputObj",
    [InputField("foo", Float), InputField("bar", NonNullType(Enum))],
)


@pytest.mark.parametrize(
    "value, expected",
    [
        (
            {"foo": 3, "bar": "HELLO"},
            _ast.ObjectValue(
                fields=[
                    _ast.ObjectField(
                        name=_ast.Name(value="foo"),
                        value=_ast.IntValue(value="3"),
                    ),
                    _ast.ObjectField(
                        name=_ast.Name(value="bar"),
                        value=_ast.EnumValue(value="HELLO"),
                    ),
                ]
            ),
        ),
        (
            {"foo": None, "bar": "HELLO"},
            _ast.ObjectValue(
                fields=[
                    _ast.ObjectField(
                        name=_ast.Name(value="foo"), value=_ast.NullValue()
                    ),
                    _ast.ObjectField(
                        name=_ast.Name(value="bar"),
                        value=_ast.EnumValue(value="HELLO"),
                    ),
                ]
            ),
        ),
        (
            {"bar": "HELLO"},
            _ast.ObjectValue(
                fields=[
                    _ast.ObjectField(
                        name=_ast.Name(value="bar"),
                        value=_ast.EnumValue(value="HELLO"),
                    )
                ]
            ),
        ),
        (42, ValueError('Value of type "MyInputObj" must be a dict')),
        ({}, ValueError('Field "bar" of type "MyInputObj" is required')),
    ],
)
def test_ast_node_from_value_with_objects(value, expected):
    if isinstance(expected, Exception):
        with pytest.raises(type(expected)) as exc_info:
            ast_node_from_value(value, InputObject)
        assert str(expected) == str(exc_info.value)
    else:
        assert ast_node_from_value(value, InputObject) == expected



# -*- coding: utf-8 -*-
""" ported from graphql-js """

import pytest
from py_gql._string_utils import parse_block_string, wrapped_lines, levenshtein


def test_parse_block_string_removes_uniform_indentation_from_a_string():
    raw_value = "\n".join(
        ["", "    Hello,", "      World!", "", "    Yours,", "      GraphQL."]
    )
    assert parse_block_string(raw_value) == "\n".join(
        ["Hello,", "  World!", "", "Yours,", "  GraphQL."]
    )


def test_parse_block_string_removes_empty_leading_and_trailing_lines():
    raw_value = "\n".join(
        [
            "",
            "",
            "    Hello,",
            "      World!",
            "",
            "    Yours,",
            "      GraphQL.",
            "",
            "",
        ]
    )

    assert parse_block_string(raw_value) == "\n".join(
        ["Hello,", "  World!", "", "Yours,", "  GraphQL."]
    )


def test_parse_block_string_removes_blank_leading_and_trailing_lines():
    raw_value = "\n".join(
        [
            "  ",
            "        ",
            "    Hello,",
            "      World!",
            "",
            "    Yours,",
            "      GraphQL.",
            "        ",
            "  ",
        ]
    )

    assert parse_block_string(raw_value) == "\n".join(
        ["Hello,", "  World!", "", "Yours,", "  GraphQL."]
    )


def test_parse_block_string_retains_indentation_from_first_line():
    raw_value = "\n".join(
        ["    Hello,", "      World!", "", "    Yours,", "      GraphQL."]
    )

    assert parse_block_string(raw_value) == "\n".join(
        ["    Hello,", "  World!", "", "Yours,", "  GraphQL."]
    )


def test_parse_block_string_does_not_alter_trailing_spaces():
    raw_value = "\n".join(
        [
            "               ",
            "    Hello,     ",
            "      World!   ",
            "               ",
            "    Yours,     ",
            "      GraphQL. ",
            "               ",
        ]
    )

    assert parse_block_string(raw_value) == "\n".join(
        [
            "Hello,     ",
            "  World!   ",
            "           ",
            "Yours,     ",
            "  GraphQL. ",
        ]
    )


def test_wrapped_lines():
    source_lines = [
        "This line is shorter and should not be wrapped.",
        "This line is long and should be wrapped at around this position.",
        "This line is longer and should be wrapped twice. This line is longer and "
        "should be wrapped twice.",
        "It should also wrap around underscores like this_token",
        "and it should also wrap around dashes like this-kind-of-token.",
    ]

    assert list(wrapped_lines(source_lines, 50)) == [
        "This line is shorter and should not be wrapped.",
        "This line is long and should be wrapped at around ",
        "this position.",
        "This line is longer and should be wrapped twice. ",
        "This line is longer and should be wrapped twice.",
        "It should also wrap around underscores like this_",
        "token",
        "and it should also wrap around dashes like this-",
        "kind-of-token.",
    ]


@pytest.mark.parametrize(
    "a,b,expected",
    [
        ("", "", 0),
        ("a", "", 1),
        ("", "a", 1),
        ("abc", "", 3),
        ("", "abc", 3),
        ("", "", 0),
        ("a", "a", 0),
        ("abc", "abc", 0),
        ("", "a", 1),
        ("a", "ab", 1),
        ("b", "ab", 1),
        ("ac", "abc", 1),
        ("abcdefg", "xabxcdxxefxgx", 6),
        ("a", "", 1),
        ("ab", "a", 1),
        ("ab", "b", 1),
        ("abc", "ac", 1),
        ("xabxcdxxefxgx", "abcdefg", 6),
        ("a", "b", 1),
        ("ab", "ac", 1),
        ("ac", "bc", 1),
        ("abc", "axc", 1),
        ("xabxcdxxefxgx", "1ab2cd34ef5g6", 6),
        ("example", "samples", 3),
        ("sturgeon", "urgently", 6),
        ("levenshtein", "frankenstein", 6),
        ("distance", "difference", 5),
        ("java was neat", "scala is great", 7),
    ],
)
def test_levenshtein(a, b, expected):
    assert levenshtein(a, b) == expected



# -*- coding: utf-8 -*-
# flake8: noqa
""" tests related to introspection queries """
from __future__ import unicode_literals

import json

from py_gql.schema import (
    Argument,
    EnumType,
    EnumValue,
    Field,
    InputField,
    InputObjectType,
    ListType,
    ObjectType,
    Schema,
    String,
)
from py_gql.utilities import introspection_query

from ._test_utils import check_execution

# Star Wars schema related tests. These are simpler than other tests down
# this file and should be easier to debug if something breaks while
# the others are supposed to cover more ground.


def test_allows_querying_the_schema_for_types(starwars_schema):
    check_execution(
        starwars_schema,
        """
    query IntrospectionTypeQuery {
        __schema {
            types { name }
        }
    }
    """,
        {
            "__schema": {
                "types": [
                    {"name": "Boolean"},
                    {"name": "Character"},
                    {"name": "Droid"},
                    {"name": "Episode"},
                    {"name": "Float"},
                    {"name": "Human"},
                    {"name": "ID"},
                    {"name": "Int"},
                    {"name": "Query"},
                    {"name": "String"},
                    {"name": "__Directive"},
                    {"name": "__DirectiveLocation"},
                    {"name": "__EnumValue"},
                    {"name": "__Field"},
                    {"name": "__InputValue"},
                    {"name": "__Schema"},
                    {"name": "__Type"},
                    {"name": "__TypeKind"},
                ]
            }
        },
        [],
    )


def test_allows_querying_the_schema_for_query_type(starwars_schema):
    check_execution(
        starwars_schema,
        """
    query IntrospectionQueryTypeQuery {
        __schema {
            queryType {
                name
            }
        }
    }
    """,
        {"__schema": {"queryType": {"name": "Query"}}},
        [],
    )


def test_allows_querying_the_schema_for_a_specific_type(starwars_schema):
    check_execution(
        starwars_schema,
        """
    query IntrospectionDroidTypeQuery {
        __type(name: "Droid") {
           name
        }
    }
    """,
        {"__type": {"name": "Droid"}},
        [],
    )


def test_allows_querying_the_schema_for_an_object_kind(starwars_schema):
    check_execution(
        starwars_schema,
        """
    query IntrospectionDroidKindQuery {
        __type(name: "Droid") {
            name
            kind
        }
    }
    """,
        {"__type": {"name": "Droid", "kind": "OBJECT"}},
        [],
    )


def test_allows_querying_the_schema_for_an_interface_kind(starwars_schema):
    check_execution(
        starwars_schema,
        """
    query IntrospectionCharacterKindQuery {
        __type(name: "Character") {
            name
           kind
        }
    }
    """,
        {"__type": {"name": "Character", "kind": "INTERFACE"}},
        [],
    )


def test_allows_querying_the_schema_for_object_fields(starwars_schema):
    check_execution(
        starwars_schema,
        """
     query IntrospectionDroidFieldsQuery {
        __type(name: "Droid") {
            name
            fields {
                name
                type {
                    name
                   kind
                }
            }
        }
    }
    """,
        {
            "__type": {
                "name": "Droid",
                "fields": [
                    {"name": "id", "type": {"name": None, "kind": "NON_NULL"}},
                    {
                        "name": "name",
                        "type": {"name": "String", "kind": "SCALAR"},
                    },
                    {"name": "friends", "type": {"name": None, "kind": "LIST"}},
                    {
                        "name": "appearsIn",
                        "type": {"name": None, "kind": "LIST"},
                    },
                    {
                        "name": "secretBackstory",
                        "type": {"name": "String", "kind": "SCALAR"},
                    },
                    {
                        "name": "primaryFunction",
                        "type": {"name": "String", "kind": "SCALAR"},
                    },
                ],
            }
        },
        [],
    )


def test_allows_querying_the_schema_for_nested_object_fields(starwars_schema):
    check_execution(
        starwars_schema,
        """
    query IntrospectionDroidNestedFieldsQuery {
        __type(name: "Droid") {
            name
            fields {
                name
                type {
                    name
                    kind
                    ofType {
                        name
                        kind
                    }
                }
            }
        }
    }
    """,
        {
            "__type": {
                "fields": [
                    {
                        "name": "id",
                        "type": {
                            "kind": "NON_NULL",
                            "name": None,
                            "ofType": {"kind": "SCALAR", "name": "String"},
                        },
                    },
                    {
                        "name": "name",
                        "type": {
                            "kind": "SCALAR",
                            "name": "String",
                            "ofType": None,
                        },
                    },
                    {
                        "name": "friends",
                        "type": {
                            "kind": "LIST",
                            "name": None,
                            "ofType": {
                                "kind": "INTERFACE",
                                "name": "Character",
                            },
                        },
                    },
                    {
                        "name": "appearsIn",
                        "type": {
                            "kind": "LIST",
                            "name": None,
                            "ofType": {"kind": "ENUM", "name": "Episode"},
                        },
                    },
                    {
                        "name": "secretBackstory",
                        "type": {
                            "kind": "SCALAR",
                            "name": "String",
                            "ofType": None,
                        },
                    },
                    {
                        "name": "primaryFunction",
                        "type": {
                            "kind": "SCALAR",
                            "name": "String",
                            "ofType": None,
                        },
                    },
                ],
                "name": "Droid",
            }
        },
        [],
    )


def test_allows_querying_the_schema_for_field_args(starwars_schema):
    check_execution(
        starwars_schema,
        """
    query IntrospectionQueryTypeQuery {
        __schema {
            queryType {
                fields {
                   name
                    args {
                        name
                        description
                        type {
                            name
                            kind
                            ofType {
                                name
                                kind
                            }
                        }
                        defaultValue
                    }
                }
            }
        }
    }
    """,
        {
            "__schema": {
                "queryType": {
                    "fields": [
                        {
                            "args": [
                                {
                                    "defaultValue": None,
                                    "description": "If omitted, returns the hero of the whole saga. "  # noqa: B950
                                    "If provided, returns the hero of that particular "
                                    "episode.",
                                    "name": "episode",
                                    "type": {
                                        "kind": "ENUM",
                                        "name": "Episode",
                                        "ofType": None,
                                    },
                                }
                            ],
                            "name": "hero",
                        },
                        {
                            "args": [
                                {
                                    "defaultValue": None,
                                    "description": "Id of the human",
                                    "name": "id",
                                    "type": {
                                        "kind": "NON_NULL",
                                        "name": None,
                                        "ofType": {
                                            "kind": "SCALAR",
                                            "name": "String",
                                        },
                                    },
                                }
                            ],
                            "name": "human",
                        },
                        {
                            "args": [
                                {
                                    "defaultValue": None,
                                    "description": "Id of the droid",
                                    "name": "id",
                                    "type": {
                                        "kind": "NON_NULL",
                                        "name": None,
                                        "ofType": {
                                            "kind": "SCALAR",
                                            "name": "String",
                                        },
                                    },
                                }
                            ],
                            "name": "droid",
                        },
                    ]
                }
            }
        },
        [],
    )


def test_allows_querying_the_schema_for_documentation(starwars_schema):
    check_execution(
        starwars_schema,
        """
    query IntrospectionDroidDescriptionQuery {
        __type(name: "Droid") {
            name
           description
        }
    }
    """,
        {
            "__type": {
                "name": "Droid",
                "description": "A mechanical creature in the Star Wars universe.",
            }
        },
        [],
    )


def test_intropsection_query():
    empty_schema = Schema(ObjectType("QueryRoot", [Field("onlyField", String)]))

    check_execution(
        empty_schema,
        introspection_query(False),
        expected_data={
            "__schema": {
                "mutationType": None,
                "subscriptionType": None,
                "queryType": {"name": "QueryRoot"},
                "types": [
                    {
                        "enumValues": None,
                        "fields": None,
                        "inputFields": None,
                        "interfaces": None,
                        "kind": "SCALAR",
                        "name": "Boolean",
                        "possibleTypes": None,
                    },
                    {
                        "enumValues": None,
                        "fields": None,
                        "inputFields": None,
                        "interfaces": None,
                        "kind": "SCALAR",
                        "name": "Float",
                        "possibleTypes": None,
                    },
                    {
                        "enumValues": None,
                        "fields": None,
                        "inputFields": None,
                        "interfaces": None,
                        "kind": "SCALAR",
                        "name": "ID",
                        "possibleTypes": None,
                    },
                    {
                        "enumValues": None,
                        "fields": None,
                        "inputFields": None,
                        "interfaces": None,
                        "kind": "SCALAR",
                        "name": "Int",
                        "possibleTypes": None,
                    },
                    {
                        "enumValues": None,
                        "fields": [
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "onlyField",
                                "type": {
                                    "kind": "SCALAR",
                                    "name": "String",
                                    "ofType": None,
                                },
                            }
                        ],
                        "inputFields": None,
                        "interfaces": [],
                        "kind": "OBJECT",
                        "name": "QueryRoot",
                        "possibleTypes": None,
                    },
                    {
                        "enumValues": None,
                        "fields": None,
                        "inputFields": None,
                        "interfaces": None,
                        "kind": "SCALAR",
                        "name": "String",
                        "possibleTypes": None,
                    },
                    {
                        "enumValues": None,
                        "fields": [
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "name",
                                "type": {
                                    "kind": "NON_NULL",
                                    "name": None,
                                    "ofType": {
                                        "kind": "SCALAR",
                                        "name": "String",
                                        "ofType": None,
                                    },
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "description",
                                "type": {
                                    "kind": "SCALAR",
                                    "name": "String",
                                    "ofType": None,
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "locations",
                                "type": {
                                    "kind": "NON_NULL",
                                    "name": None,
                                    "ofType": {
                                        "kind": "LIST",
                                        "name": None,
                                        "ofType": {
                                            "kind": "NON_NULL",
                                            "name": None,
                                            "ofType": {
                                                "kind": "ENUM",
                                                "name": "__DirectiveLocation",
                                                "ofType": None,
                                            },
                                        },
                                    },
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "args",
                                "type": {
                                    "kind": "NON_NULL",
                                    "name": None,
                                    "ofType": {
                                        "kind": "LIST",
                                        "name": None,
                                        "ofType": {
                                            "kind": "NON_NULL",
                                            "name": None,
                                            "ofType": {
                                                "kind": "OBJECT",
                                                "name": "__InputValue",
                                                "ofType": None,
                                            },
                                        },
                                    },
                                },
                            },
                        ],
                        "inputFields": None,
                        "interfaces": [],
                        "kind": "OBJECT",
                        "name": "__Directive",
                        "possibleTypes": None,
                    },
                    {
                        "enumValues": [
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "QUERY",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "MUTATION",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "SUBSCRIPTION",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "FIELD",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "FRAGMENT_DEFINITION",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "FRAGMENT_SPREAD",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "INLINE_FRAGMENT",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "SCHEMA",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "SCALAR",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "OBJECT",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "FIELD_DEFINITION",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "ARGUMENT_DEFINITION",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "INTERFACE",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "UNION",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "ENUM",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "ENUM_VALUE",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "INPUT_OBJECT",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "INPUT_FIELD_DEFINITION",
                            },
                        ],
                        "fields": None,
                        "inputFields": None,
                        "interfaces": None,
                        "kind": "ENUM",
                        "name": "__DirectiveLocation",
                        "possibleTypes": None,
                    },
                    {
                        "enumValues": None,
                        "fields": [
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "name",
                                "type": {
                                    "kind": "NON_NULL",
                                    "name": None,
                                    "ofType": {
                                        "kind": "SCALAR",
                                        "name": "String",
                                        "ofType": None,
                                    },
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "description",
                                "type": {
                                    "kind": "SCALAR",
                                    "name": "String",
                                    "ofType": None,
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "isDeprecated",
                                "type": {
                                    "kind": "NON_NULL",
                                    "name": None,
                                    "ofType": {
                                        "kind": "SCALAR",
                                        "name": "Boolean",
                                        "ofType": None,
                                    },
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "deprecationReason",
                                "type": {
                                    "kind": "SCALAR",
                                    "name": "String",
                                    "ofType": None,
                                },
                            },
                        ],
                        "inputFields": None,
                        "interfaces": [],
                        "kind": "OBJECT",
                        "name": "__EnumValue",
                        "possibleTypes": None,
                    },
                    {
                        "enumValues": None,
                        "fields": [
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "name",
                                "type": {
                                    "kind": "NON_NULL",
                                    "name": None,
                                    "ofType": {
                                        "kind": "SCALAR",
                                        "name": "String",
                                        "ofType": None,
                                    },
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "description",
                                "type": {
                                    "kind": "SCALAR",
                                    "name": "String",
                                    "ofType": None,
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "args",
                                "type": {
                                    "kind": "NON_NULL",
                                    "name": None,
                                    "ofType": {
                                        "kind": "LIST",
                                        "name": None,
                                        "ofType": {
                                            "kind": "NON_NULL",
                                            "name": None,
                                            "ofType": {
                                                "kind": "OBJECT",
                                                "name": "__InputValue",
                                                "ofType": None,
                                            },
                                        },
                                    },
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "type",
                                "type": {
                                    "kind": "NON_NULL",
                                    "name": None,
                                    "ofType": {
                                        "kind": "OBJECT",
                                        "name": "__Type",
                                        "ofType": None,
                                    },
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "isDeprecated",
                                "type": {
                                    "kind": "NON_NULL",
                                    "name": None,
                                    "ofType": {
                                        "kind": "SCALAR",
                                        "name": "Boolean",
                                        "ofType": None,
                                    },
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "deprecationReason",
                                "type": {
                                    "kind": "SCALAR",
                                    "name": "String",
                                    "ofType": None,
                                },
                            },
                        ],
                        "inputFields": None,
                        "interfaces": [],
                        "kind": "OBJECT",
                        "name": "__Field",
                        "possibleTypes": None,
                    },
                    {
                        "enumValues": None,
                        "fields": [
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "name",
                                "type": {
                                    "kind": "NON_NULL",
                                    "name": None,
                                    "ofType": {
                                        "kind": "SCALAR",
                                        "name": "String",
                                        "ofType": None,
                                    },
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "description",
                                "type": {
                                    "kind": "SCALAR",
                                    "name": "String",
                                    "ofType": None,
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "type",
                                "type": {
                                    "kind": "NON_NULL",
                                    "name": None,
                                    "ofType": {
                                        "kind": "OBJECT",
                                        "name": "__Type",
                                        "ofType": None,
                                    },
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "defaultValue",
                                "type": {
                                    "kind": "SCALAR",
                                    "name": "String",
                                    "ofType": None,
                                },
                            },
                        ],
                        "inputFields": None,
                        "interfaces": [],
                        "kind": "OBJECT",
                        "name": "__InputValue",
                        "possibleTypes": None,
                    },
                    {
                        "enumValues": None,
                        "fields": [
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "types",
                                "type": {
                                    "kind": "NON_NULL",
                                    "name": None,
                                    "ofType": {
                                        "kind": "LIST",
                                        "name": None,
                                        "ofType": {
                                            "kind": "NON_NULL",
                                            "name": None,
                                            "ofType": {
                                                "kind": "OBJECT",
                                                "name": "__Type",
                                                "ofType": None,
                                            },
                                        },
                                    },
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "queryType",
                                "type": {
                                    "kind": "NON_NULL",
                                    "name": None,
                                    "ofType": {
                                        "kind": "OBJECT",
                                        "name": "__Type",
                                        "ofType": None,
                                    },
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "mutationType",
                                "type": {
                                    "kind": "OBJECT",
                                    "name": "__Type",
                                    "ofType": None,
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "subscriptionType",
                                "type": {
                                    "kind": "OBJECT",
                                    "name": "__Type",
                                    "ofType": None,
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "directives",
                                "type": {
                                    "kind": "NON_NULL",
                                    "name": None,
                                    "ofType": {
                                        "kind": "LIST",
                                        "name": None,
                                        "ofType": {
                                            "kind": "NON_NULL",
                                            "name": None,
                                            "ofType": {
                                                "kind": "OBJECT",
                                                "name": "__Directive",
                                                "ofType": None,
                                            },
                                        },
                                    },
                                },
                            },
                        ],
                        "inputFields": None,
                        "interfaces": [],
                        "kind": "OBJECT",
                        "name": "__Schema",
                        "possibleTypes": None,
                    },
                    {
                        "enumValues": None,
                        "fields": [
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "kind",
                                "type": {
                                    "kind": "NON_NULL",
                                    "name": None,
                                    "ofType": {
                                        "kind": "ENUM",
                                        "name": "__TypeKind",
                                        "ofType": None,
                                    },
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "name",
                                "type": {
                                    "kind": "SCALAR",
                                    "name": "String",
                                    "ofType": None,
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "description",
                                "type": {
                                    "kind": "SCALAR",
                                    "name": "String",
                                    "ofType": None,
                                },
                            },
                            {
                                "args": [
                                    {
                                        "defaultValue": "false",
                                        "name": "includeDeprecated",
                                        "type": {
                                            "kind": "SCALAR",
                                            "name": "Boolean",
                                            "ofType": None,
                                        },
                                    }
                                ],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "fields",
                                "type": {
                                    "kind": "LIST",
                                    "name": None,
                                    "ofType": {
                                        "kind": "NON_NULL",
                                        "name": None,
                                        "ofType": {
                                            "kind": "OBJECT",
                                            "name": "__Field",
                                            "ofType": None,
                                        },
                                    },
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "interfaces",
                                "type": {
                                    "kind": "LIST",
                                    "name": None,
                                    "ofType": {
                                        "kind": "NON_NULL",
                                        "name": None,
                                        "ofType": {
                                            "kind": "OBJECT",
                                            "name": "__Type",
                                            "ofType": None,
                                        },
                                    },
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "possibleTypes",
                                "type": {
                                    "kind": "LIST",
                                    "name": None,
                                    "ofType": {
                                        "kind": "NON_NULL",
                                        "name": None,
                                        "ofType": {
                                            "kind": "OBJECT",
                                            "name": "__Type",
                                            "ofType": None,
                                        },
                                    },
                                },
                            },
                            {
                                "args": [
                                    {
                                        "defaultValue": "false",
                                        "name": "includeDeprecated",
                                        "type": {
                                            "kind": "SCALAR",
                                            "name": "Boolean",
                                            "ofType": None,
                                        },
                                    }
                                ],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "enumValues",
                                "type": {
                                    "kind": "LIST",
                                    "name": None,
                                    "ofType": {
                                        "kind": "NON_NULL",
                                        "name": None,
                                        "ofType": {
                                            "kind": "OBJECT",
                                            "name": "__EnumValue",
                                            "ofType": None,
                                        },
                                    },
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "inputFields",
                                "type": {
                                    "kind": "LIST",
                                    "name": None,
                                    "ofType": {
                                        "kind": "NON_NULL",
                                        "name": None,
                                        "ofType": {
                                            "kind": "OBJECT",
                                            "name": "__InputValue",
                                            "ofType": None,
                                        },
                                    },
                                },
                            },
                            {
                                "args": [],
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "ofType",
                                "type": {
                                    "kind": "OBJECT",
                                    "name": "__Type",
                                    "ofType": None,
                                },
                            },
                        ],
                        "inputFields": None,
                        "interfaces": [],
                        "kind": "OBJECT",
                        "name": "__Type",
                        "possibleTypes": None,
                    },
                    {
                        "enumValues": [
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "SCALAR",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "OBJECT",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "INTERFACE",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "UNION",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "ENUM",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "INPUT_OBJECT",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "LIST",
                            },
                            {
                                "deprecationReason": None,
                                "isDeprecated": False,
                                "name": "NON_NULL",
                            },
                        ],
                        "fields": None,
                        "inputFields": None,
                        "interfaces": None,
                        "kind": "ENUM",
                        "name": "__TypeKind",
                        "possibleTypes": None,
                    },
                ],
                "directives": [
                    {
                        "name": "deprecated",
                        "locations": ["FIELD_DEFINITION", "ENUM_VALUE"],
                        "args": [
                            {
                                "defaultValue": '"No longer supported"',
                                "name": "reason",
                                "type": {
                                    "kind": "SCALAR",
                                    "name": "String",
                                    "ofType": None,
                                },
                            }
                        ],
                    },
                    {
                        "name": "include",
                        "locations": [
                            "FIELD",
                            "FRAGMENT_SPREAD",
                            "INLINE_FRAGMENT",
                        ],
                        "args": [
                            {
                                "defaultValue": None,
                                "name": "if",
                                "type": {
                                    "kind": "NON_NULL",
                                    "name": None,
                                    "ofType": {
                                        "kind": "SCALAR",
                                        "name": "Boolean",
                                        "ofType": None,
                                    },
                                },
                            }
                        ],
                    },
                    {
                        "name": "skip",
                        "locations": [
                            "FIELD",
                            "FRAGMENT_SPREAD",
                            "INLINE_FRAGMENT",
                        ],
                        "args": [
                            {
                                "defaultValue": None,
                                "name": "if",
                                "type": {
                                    "kind": "NON_NULL",
                                    "name": None,
                                    "ofType": {
                                        "kind": "SCALAR",
                                        "name": "Boolean",
                                        "ofType": None,
                                    },
                                },
                            }
                        ],
                    },
                ],
            }
        },
        expected_errors=[],
    )


def test_intropsection_on_input_object():
    test_input = InputObjectType(
        "TestInputObject",
        [
            InputField("a", String, default_value="tes\t de\fault"),
            InputField("b", ListType(String)),
            InputField("c", String, default_value=None),
        ],
    )

    test_type = ObjectType(
        "TestType",
        [
            Field(
                "field",
                String,
                [Argument("complex", test_input)],
                resolve=lambda _, args, *__: json.dumps(args.get("complex")),
            )
        ],
    )

    schema = Schema(test_type)

    check_execution(
        schema,
        """
    {
        __type(name: "TestInputObject") {
            kind
            name
            inputFields {
                name
                type { ...TypeRef }
                defaultValue
            }
        }
    }

    fragment TypeRef on __Type {
        kind
        name
        ofType {
            kind
            name
            ofType {
                kind
                name
                ofType {
                    kind
                    name
                }
            }
        }
    }
    """,
        expected_data={
            "__type": {
                "inputFields": [
                    {
                        "defaultValue": '"tes\t de\x0cault"',
                        "name": "a",
                        "type": {
                            "kind": "SCALAR",
                            "name": "String",
                            "ofType": None,
                        },
                    },
                    {
                        "defaultValue": None,
                        "name": "b",
                        "type": {
                            "kind": "LIST",
                            "name": None,
                            "ofType": {
                                "kind": "SCALAR",
                                "name": "String",
                                "ofType": None,
                            },
                        },
                    },
                    {
                        "defaultValue": "null",
                        "name": "c",
                        "type": {
                            "kind": "SCALAR",
                            "name": "String",
                            "ofType": None,
                        },
                    },
                ],
                "kind": "INPUT_OBJECT",
                "name": "TestInputObject",
            }
        },
        expected_errors=[],
    )


def test_it_supports_the_type_root_field():
    test_type = ObjectType("TestType", [Field("testField", String)])
    schema = Schema(test_type)
    check_execution(
        schema,
        """
    {
        __type(name: "TestType") {
            name
        }
    }
    """,
        expected_data={"__type": {"name": "TestType"}},
        expected_errors=[],
    )


def test_it_identifies_deprecated_fields():
    test_type = ObjectType(
        "TestType",
        [
            Field("nonDeprecated", String),
            Field("deprecated", String, deprecation_reason="Removed in 1.0"),
        ],
    )
    schema = Schema(test_type)

    check_execution(
        schema,
        """
    {
        __type(name: "TestType") {
            name
            fields(includeDeprecated: true) {
                name
                isDeprecated,
               deprecationReason
            }
        }
    }
    """,
        expected_data={
            "__type": {
                "name": "TestType",
                "fields": [
                    {
                        "name": "nonDeprecated",
                        "isDeprecated": False,
                        "deprecationReason": None,
                    },
                    {
                        "name": "deprecated",
                        "isDeprecated": True,
                        "deprecationReason": "Removed in 1.0",
                    },
                ],
            }
        },
        expected_errors=[],
    )


def test_it_respects_the_include_deprecated_parameter_for_fields():
    test_type = ObjectType(
        "TestType",
        [
            Field("nonDeprecated", String),
            Field("deprecated", String, deprecation_reason="Removed in 1.0"),
        ],
    )
    schema = Schema(test_type)

    check_execution(
        schema,
        """
    {
        __type(name: "TestType") {
            name
            fields {
                name
                isDeprecated,
               deprecationReason
            }
        }
    }
    """,
        expected_data={
            "__type": {
                "name": "TestType",
                "fields": [
                    {
                        "name": "nonDeprecated",
                        "isDeprecated": False,
                        "deprecationReason": None,
                    }
                ],
            }
        },
        expected_errors=[],
    )


def test_it_identifies_deprecated_enum_values():
    test_enum = EnumType(
        "TestEnum",
        [
            EnumValue("NONDEPRECATED", 0),
            EnumValue("DEPRECATED", 1, "Removed in 1.0"),
            EnumValue("ALSONONDEPRECATED", 2),
        ],
    )
    test_type = ObjectType("TestType", [Field("testEnum", test_enum)])
    schema = Schema(test_type)

    check_execution(
        schema,
        """
    {
        __type(name: "TestEnum") {
            name
            enumValues(includeDeprecated: true) {
                name
                isDeprecated,
                deprecationReason
            }
        }
    }
    """,
        {
            "__type": {
                "enumValues": [
                    {
                        "deprecationReason": None,
                        "isDeprecated": False,
                        "name": "NONDEPRECATED",
                    },
                    {
                        "deprecationReason": "Removed in 1.0",
                        "isDeprecated": True,
                        "name": "DEPRECATED",
                    },
                    {
                        "deprecationReason": None,
                        "isDeprecated": False,
                        "name": "ALSONONDEPRECATED",
                    },
                ],
                "name": "TestEnum",
            }
        },
        [],
    )


def test_it_respects_the_include_deprecated_parameter_for_enum_values():
    test_enum = EnumType(
        "TestEnum",
        [
            EnumValue("NONDEPRECATED", 0),
            EnumValue("DEPRECATED", 1, "Removed in 1.0"),
            EnumValue("ALSONONDEPRECATED", 2),
        ],
    )
    test_type = ObjectType("TestType", [Field("testEnum", test_enum)])
    schema = Schema(test_type)

    check_execution(
        schema,
        """
    {
        __type(name: "TestEnum") {
            name
            enumValues(includeDeprecated: false) {
                name
                isDeprecated,
                deprecationReason
            }
        }
    }
    """,
        {
            "__type": {
                "enumValues": [
                    {
                        "deprecationReason": None,
                        "isDeprecated": False,
                        "name": "NONDEPRECATED",
                    },
                    {
                        "deprecationReason": None,
                        "isDeprecated": False,
                        "name": "ALSONONDEPRECATED",
                    },
                ],
                "name": "TestEnum",
            }
        },
        [],
    )


def test_it_fails_as_expected_on_the_type_root_field_without_an_arg():
    test_type = ObjectType("TestType", [Field("testField", String)])
    schema = Schema(test_type)
    check_execution(
        schema,
        "{ __type { name } }",
        expected_data={"__type": None},
        expected_errors=[
            (
                'Argument "name" of required type "String!" was not provided',
                (2, 17),
                "__type",
            )
        ],
    )


def test_it_exposes_descriptions_on_types_and_fields():
    schema = Schema(ObjectType("QueryRoot", [Field("onlyField", String)]))
    check_execution(
        schema,
        """
    {
        schemaType: __type(name: "__Schema") {
            name,
            description,
            fields { name, description }
        }
    }
    """,
        {
            "schemaType": {
                "name": "__Schema",
                "description": "A GraphQL Schema defines the capabilities of a "
                "GraphQL server. It exposes all available types and "
                "directives on the server, as well as the entry "
                "points for query, mutation, "
                "and subscription operations.",
                "fields": [
                    {
                        "name": "types",
                        "description": "A list of all types supported by this server.",
                    },
                    {
                        "name": "queryType",
                        "description": "The type that query operations will be rooted at.",
                    },
                    {
                        "name": "mutationType",
                        "description": "If this server supports mutation, the type that "
                        "mutation operations will be rooted at.",
                    },
                    {
                        "name": "subscriptionType",
                        "description": "If this server supports subscription, the type "
                        "that subscription operations will be rooted at.",
                    },
                    {
                        "name": "directives",
                        "description": "A list of all directives supported by this server.",
                    },
                ],
            }
        },
        [],
    )


def test_it_exposes_descriptions_on_enums():
    schema = Schema(ObjectType("QueryRoot", [Field("onlyField", String)]))
    check_execution(
        schema,
        """
    {
        typeKindType: __type(name: "__TypeKind") {
            name,
            description,
            enumValues {
                name,
                description
            }
        }
    }
    """,
        {
            "typeKindType": {
                "description": "An enum describing what kind of type a given `__Type` is.",
                "enumValues": [
                    {
                        "description": "Indicates this type is a scalar.",
                        "name": "SCALAR",
                    },
                    {
                        "description": "Indicates this type is an object. `fields` and "
                        "`interfaces` are valid fields.",
                        "name": "OBJECT",
                    },
                    {
                        "description": "Indicates this type is an interface. `fields` and "
                        "`possibleTypes` are valid fields.",
                        "name": "INTERFACE",
                    },
                    {
                        "description": "Indicates this type is a union. `possibleTypes` is a "
                        "valid field.",
                        "name": "UNION",
                    },
                    {
                        "description": "Indicates this type is an enum. `enumValues` is a "
                        "valid field.",
                        "name": "ENUM",
                    },
                    {
                        "description": "Indicates this type is an input object. `inputFields` "
                        "is a valid field.",
                        "name": "INPUT_OBJECT",
                    },
                    {
                        "description": "Indicates this type is a list. `ofType` is a valid field.",
                        "name": "LIST",
                    },
                    {
                        "description": "Indicates this type is a non-null. `ofType` is a "
                        "valid field.",
                        "name": "NON_NULL",
                    },
                ],
                "name": "__TypeKind",
            }
        },
        [],
    )



# -*- coding: utf-8 -*-
"""
"""

import functools as ft

from .. import types as _types
from ..._utils import lazy
from ...exc import ExtensionError, SDLError
from ...lang import ast as _ast
from ...utilities import directive_arguments, value_from_ast
from ..directives import SPECIFIED_DIRECTIVES, DeprecatedDirective
from ..introspection import INTROPSPECTION_TYPES
from ..scalars import SPECIFIED_SCALAR_TYPES, default_scalar


def _deprecation_reason(node):
    # type: (Union[_ast.FieldDefinition, _ast.EnumTypeDefinition]) -> Optional[str]
    args = directive_arguments(DeprecatedDirective, node, {})
    return args.get("reason", None) if args else None


def _desc(node):
    # type: (Union[_ast.TypeDefinition, _ast.FieldDefinition, \
    # _ast.EnumValueDefinition, _ast.DirectiveDefinition]) -> Optional[str]
    return node.description.value if node.description else None


_DEFAULT_TYPES_MAP = {
    t.name: t for t in (SPECIFIED_SCALAR_TYPES + INTROPSPECTION_TYPES)
}


_EXPECTED_EXTENSIONS = {
    _types.ObjectType: _ast.ObjectTypeExtension,
    _types.InterfaceType: _ast.InterfaceTypeExtension,
    _types.EnumType: _ast.EnumTypeExtension,
    _types.UnionType: _ast.UnionTypeExtension,
    _types.InputObjectType: _ast.InputObjectTypeExtension,
    _types.ScalarType: _ast.ScalarTypeExtension,
}


def _assert_correct_extension(type_cls, extension_node):
    try:
        expected_node_cls = _EXPECTED_EXTENSIONS[type_cls]
    except KeyError:
        pass
    else:
        if not isinstance(extension_node, expected_node_cls):
            raise ExtensionError(
                "Expected %s when extending %s but got %s"
                % (
                    expected_node_cls.__name__,
                    type_cls.__name__,
                    type(extension_node).__name__,
                ),
                [extension_node],
            )


class TypesBuilder(object):
    """ Build and extend type definitions from AST nodes.

    Warning:
        Specified types (scalars, introspection and directives) will not be
        built or extended.

    Args:
        type_defs (Mapping[str, _ast.TypeDefinition]):
            Type definitions by name
        directive_defs (Mapping[str, _ast.DirectiveDefinition]):
            Directive definitions by name
        type_extensions (Mapping[str, List[_ast.TypeExtensions]]):
            Type extensions by extend type name
        additional_types (Mapping[str, _types.NamedType]):
            Known type that should not be built when referenced
    """

    __slots__ = (
        "_type_defs",
        "_directive_defs",
        "_extensions",
        "_cache",
        "_extended_cache",
    )

    def __init__(
        self,
        type_defs,
        directive_defs,
        type_extensions=None,
        additional_types=None,
    ):
        self._type_defs = type_defs
        self._directive_defs = directive_defs
        self._extensions = type_extensions or {}
        self._cache = dict(_DEFAULT_TYPES_MAP)
        if additional_types:
            self._cache.update(additional_types)
        self._extended_cache = {}

    def build_type(self, type_node):
        # type: (Union[_ast.Type, _ast.TypeSystemDefinition]) -> _types.Type
        if isinstance(type_node, _ast.NonNullType):
            return _types.NonNullType(
                self.build_type(type_node.type), node=type_node
            )

        if isinstance(type_node, _ast.ListType):
            return _types.ListType(
                self.build_type(type_node.type), node=type_node
            )

        type_name = type_node.name.value

        try:
            return self._cache[type_name]
        except KeyError:
            if isinstance(type_node, _ast.NamedType):
                type_def = self._type_defs.get(type_name, None)
                if type_def is None:
                    raise SDLError(
                        "Type %s not found in document" % type_name, [type_node]
                    )
            else:
                type_def = type_node

            self._cache[type_name] = built = self._type_from_definition(
                type_def
            )
            return built

    def build_directive(self, directive_def):
        # type: (_ast.DirectiveDefinition) -> _types.Directive
        return _types.Directive(
            name=directive_def.name.value,
            description=_desc(directive_def),
            locations=[loc.value for loc in directive_def.locations],
            args=(
                [
                    self._build_input_value(arg, _types.Argument)
                    for arg in directive_def.arguments
                ]
                if directive_def.arguments
                else None
            ),
            node=directive_def,
        )

    def _type_from_definition(self, type_def):
        # type: (_ast.TypeSystemDefinition) -> _types.Type
        if isinstance(type_def, _ast.ObjectTypeDefinition):
            return self._build_object_type(type_def)
        if isinstance(type_def, _ast.InterfaceTypeDefinition):
            return self._build_interface_type(type_def)
        if isinstance(type_def, _ast.EnumTypeDefinition):
            return self._build_enum_type(type_def)
        if isinstance(type_def, _ast.UnionTypeDefinition):
            return self._build_union_type(type_def)
        if isinstance(type_def, _ast.ScalarTypeDefinition):
            return self._build_scalar_type(type_def)
        if isinstance(type_def, _ast.InputObjectTypeDefinition):
            return self._build_input_object_type(type_def)
        raise TypeError(type(type_def))

    def _build_object_type(self, type_def):
        # type: (_ast.ObjectTypeDefinition) -> _types.ObjectType
        return _types.ObjectType(
            name=type_def.name.value,
            description=_desc(type_def),
            fields=[
                self._build_fields(field_node) for field_node in type_def.fields
            ],
            interfaces=(
                [self.build_type(iface) for iface in type_def.interfaces]
                if type_def.interfaces
                else None
            ),
            nodes=[type_def],
        )

    def _build_interface_type(self, type_def):
        # type: (_ast.InterfaceTypeDefinition) -> _types.InterfaceType
        return _types.InterfaceType(
            name=type_def.name.value,
            description=_desc(type_def),
            fields=[
                self._build_fields(field_node) for field_node in type_def.fields
            ],
            nodes=[type_def],
        )

    def _build_fields(self, field_def):
        # type: (_ast.FieldDefinition) -> _types.Field
        return _types.Field(
            field_def.name.value,
            ft.partial(self.build_type, field_def.type),
            description=_desc(field_def),
            args=(
                [
                    self._build_input_value(arg, _types.Argument)
                    for arg in field_def.arguments
                ]
                if field_def.arguments
                else None
            ),
            deprecation_reason=_deprecation_reason(field_def),
            node=field_def,
        )

    def _build_enum_type(self, type_def):
        # type: (_ast.EnumTypeDefinition) -> _types.EnumType
        return _types.EnumType(
            name=type_def.name.value,
            description=_desc(type_def),
            values=[self._build_enum_value(v) for v in type_def.values],
            nodes=[type_def],
        )

    def _build_enum_value(self, node):
        # type: (_ast.EnumValueDefinition) -> _types.EnumValue
        return _types.EnumValue(
            name=node.name.value,
            description=_desc(node),
            value=node.name.value,
            deprecation_reason=_deprecation_reason(node),
            node=node,
        )

    def _build_union_type(self, type_def):
        # type: (_ast.UnionTypeDefinition) -> _types.UnionType
        return _types.UnionType(
            name=type_def.name.value,
            description=_desc(type_def),
            types=[self.build_type(type_) for type_ in type_def.types],
            nodes=[type_def],
        )

    def _build_scalar_type(self, type_def):
        # type: (_ast.ScalarTypeDefinition) -> _types.ScalarType
        return default_scalar(
            name=type_def.name.value,
            description=_desc(type_def),
            nodes=[type_def],
        )

    def _build_input_object_type(self, type_def):
        # type: (_ast.InputObjectTypeDefinition) -> _types.InputObjectType
        return _types.InputObjectType(
            name=type_def.name.value,
            description=_desc(type_def),
            fields=[
                self._build_input_value(field_node, _types.InputField)
                for field_node in type_def.fields
            ],
            nodes=[type_def],
        )

    def _build_input_value(
        self,
        input_value_def,  # type: _ast.InputValueDefinition
        input_value_cls,  # type: Type
    ):
        # type: (...) -> Union[_types.Argument, _types.InputField]
        type_ = self.build_type(input_value_def.type)
        kwargs = dict(description=_desc(input_value_def), node=input_value_def)
        if input_value_def.default_value is not None:
            kwargs["default_value"] = value_from_ast(
                input_value_def.default_value, lazy(type_)
            )
        return input_value_cls(input_value_def.name.value, type_, **kwargs)

    def extend_type(self, type_):
        # type: (_types.Type) -> _types.Type
        cache = self._extended_cache

        if isinstance(type_, _types.ListType):
            return _types.ListType(self.extend_type(type_.type))

        if isinstance(type_, _types.NonNullType):
            return _types.NonNullType(self.extend_type(type_.type))

        name = type_.name

        if name in _DEFAULT_TYPES_MAP:
            return type_

        try:
            return cache[name]
        except KeyError:
            if isinstance(type_, _types.ObjectType):
                cache[name] = extended = self._extend_object_type(type_)
            elif isinstance(type_, _types.InterfaceType):
                cache[name] = extended = self._extend_interface_type(type_)
            elif isinstance(type_, _types.EnumType):
                cache[name] = extended = self._extend_enum_type(type_)
            elif isinstance(type_, _types.UnionType):
                cache[name] = extended = self._extend_union_type(type_)
            elif isinstance(type_, _types.InputObjectType):
                cache[name] = extended = self._extend_input_object_type(type_)
            elif isinstance(type_, _types.ScalarType):
                cache[name] = extended = self._extend_scalar_type(type_)
            else:
                raise TypeError(type(type_))

            return extended

    def extend_directive(self, directive):
        # type: (_types.Directive) -> _types.Directive
        if directive in SPECIFIED_DIRECTIVES:
            return directive

        return _types.Directive(
            directive.name,
            description=directive.description,
            locations=directive.locations,
            args=[self._extend_argument(a) for a in directive.args],
            node=directive.node,
        )

    def _extend_object_type(self, object_type):
        # type: (_types.ObjectType) -> _types.ObjectType
        name = object_type.name
        extensions = self._extensions.get(name, [])

        for extension_node in extensions:
            _assert_correct_extension(_types.ObjectType, extension_node)

        return _types.ObjectType(
            name,
            description=object_type.description,
            fields=self._extend_object_fields(object_type),
            interfaces=self._extend_object_interfaces(object_type),
            nodes=object_type.nodes + extensions,
        )

    def _extend_object_interfaces(self, object_type):
        # type: (_types.ObjectType) -> List[_types.NamedType
        iface_names = set(i.name for i in object_type.interfaces)
        ifaces = [self.extend_type(iface) for iface in object_type.interfaces]

        for extension in self._extensions.get(object_type.name, []):
            for ext_iface in extension.interfaces:
                if ext_iface.name.value in iface_names:
                    raise ExtensionError(
                        'Interface "%s" already implemented for type "%s"'
                        % (ext_iface.name.value, object_type.name),
                        [ext_iface],
                    )
                iface_names.add(ext_iface.name.value)
                ifaces.append(self.extend_type(self.build_type(ext_iface)))

        return ifaces

    def _extend_object_fields(self, composite_type):
        # type: (Union[_types.ObjectType, _types.InterfaceType]) -> List[_types.Field]
        field_names = set(f.name for f in composite_type.fields)
        fields = [
            _types.Field(
                f.name,
                self.extend_type(f.type),
                description=f.description,
                deprecation_reason=f.deprecation_reason,
                args=[self._extend_argument(a) for a in f.args],
                resolve=f.resolve,
                node=f.node,
            )
            for f in composite_type.fields
        ]

        for extension in self._extensions.get(composite_type.name, []):
            for ext_field in extension.fields:
                if ext_field.name.value in field_names:
                    raise ExtensionError(
                        'Found duplicate field "%s" when extending %s "%s"'
                        % (
                            ext_field.name.value,
                            "type"
                            if isinstance(composite_type, _types.ObjectType)
                            else "interface",
                            composite_type.name,
                        ),
                        [ext_field],
                    )
                field_names.add(ext_field.name.value)
                fields.append(self._build_fields(ext_field))

        return fields

    def _extend_interface_type(self, interface_type):
        # type: (_types.InterfaceType) -> _types.InterfaceType
        name = interface_type.name
        extensions = self._extensions.get(name, [])

        for extension_node in extensions:
            _assert_correct_extension(_types.InterfaceType, extension_node)

        return _types.InterfaceType(
            name,
            description=interface_type.description,
            fields=self._extend_object_fields(interface_type),
            nodes=interface_type.nodes + extensions,
        )

    def _extend_enum_type(self, enum_type):
        # type: (_types.EnumType) -> EnumType
        name = enum_type.name
        extensions = self._extensions.get(name, [])

        values = enum_type.values[:]
        value_names = set(ev.name for ev in values)

        for extension_node in extensions:
            _assert_correct_extension(_types.EnumType, extension_node)
            for value in extension_node.values:
                if value.name.value in value_names:
                    raise ExtensionError(
                        'Found duplicate enum value "%s" when extending EnumType "%s"'
                        % (value.name.value, name),
                        [value],
                    )
                values.append(self._build_enum_value(value))
                value_names.add(value.name.value)

        return _types.EnumType(
            name,
            description=enum_type.description,
            values=values,
            nodes=enum_type.nodes + extensions,
        )

    def _extend_union_type(self, union_type):
        # type: (_types.UnionType) -> UnionType
        name = union_type.name
        extensions = self._extensions.get(name, [])

        member_names = set(t.name for t in union_type.types)
        member_types = [self.extend_type(t) for t in union_type.types]

        for extension_node in extensions:
            _assert_correct_extension(_types.UnionType, extension_node)
            for type_def in extension_node.types:
                if type_def.name.value in member_names:
                    raise ExtensionError(
                        'Found duplicate member type "%s" when extending UnionType "%s"'
                        % (type_def.name.value, name),
                        [type_def],
                    )
                member_types.append(self.extend_type(self.build_type(type_def)))
                member_names.add(type_def.name.value)

        return _types.UnionType(
            name, types=member_types, nodes=union_type.nodes + extensions
        )

    def _extend_input_object_type(self, input_object_type):
        # type: (_types.InputObjectType) -> InputObjectType
        name = input_object_type.name
        extensions = self._extensions.get(name, [])

        field_names = set(f.name for f in input_object_type.fields)
        fields = [
            _types.InputField(
                f.name,
                self.extend_type(f.type),
                default_value=f._default_value,
                description=f.description,
                node=f.node,
            )
            for f in input_object_type.fields
        ]

        for extension_node in extensions:
            _assert_correct_extension(_types.InputObjectType, extension_node)
            for ext_field in extension_node.fields:
                if ext_field.name.value in field_names:
                    raise ExtensionError(
                        'Found duplicate field "%s" when extending input object "%s"'
                        % (ext_field.name.value, name),
                        [ext_field],
                    )
                field_names.add(ext_field.name.value)
                fields.append(
                    self._build_input_value(ext_field, _types.InputField)
                )

        return _types.InputObjectType(
            name,
            description=input_object_type.description,
            fields=fields,
            nodes=input_object_type.nodes + extensions,
        )

    def _extend_scalar_type(self, scalar_type):
        # type: (_types.ScalarType) -> ScalarType
        name = scalar_type.name
        extensions = self._extensions.get(name, [])

        for extension_node in extensions:
            _assert_correct_extension(_types.ScalarType, extension_node)

        return _types.ScalarType(
            name,
            description=scalar_type.description,
            serialize=scalar_type._serialize,
            parse=scalar_type._parse,
            parse_literal=scalar_type._parse_literal,
            nodes=scalar_type.nodes + extensions,
        )

    def _extend_argument(self, argument):
        # type: (_types.Argument) -> Argument
        return _types.Argument(
            argument.name,
            self.extend_type(argument.type),
            default_value=argument._default_value,
            description=argument.description,
            node=argument.node,
        )



# -*- coding: utf-8 -*-

import pytest

from py_gql._graphql import graphql
from py_gql._string_utils import stringify_path
from py_gql.execution.executors import ThreadPoolExecutor
from py_gql.execution.middleware import apply_middlewares

from ._test_utils import TESTED_EXECUTORS


def test_apply_middlewares():
    def a(n, *a, **k):
        k["steps"].append("> a")
        return n(*a, **k)

    def b(n, *a, **k):
        k["steps"].append("> b")
        yield n(*a, **k)
        k["steps"].append("< b")

    def c(n, *a, **k):
        k["steps"].append("> c")
        return n(*a, **k)

    def source(*a, **k):
        return (a, k)

    assert apply_middlewares(source, [a, b, c])(42, bar=42, steps=[]) == (
        (42,),
        {"bar": 42, "steps": ["> a", "> b", "> c", "< b"]},
    )


def test_apply_middlewares_noop():
    def source(*a, **k):
        return (a, k)

    assert apply_middlewares(source, []) is source


def test_apply_middlewares_no_yield():
    def never_yields(n, *a, **k):
        # pylint: disable = using-constant-test
        if False:
            yield n(*a, **k)

    def source(*a, **k):
        return (a, k)

    with pytest.raises(RuntimeError):
        apply_middlewares(source, [never_yields])(42)


HERO_QUERY = """
query HeroNameAndFriendsQuery {
    hero {
        id
        name
        friends {
            name
        }
    }
}
"""


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_function_middleware(exe_cls, exe_kwargs, starwars_schema):
    log = []

    def path_collector_one_way(next_, root, args, context, info):
        log.append("> %s" % stringify_path(info.path))
        return next_(root, args, context, info)

    with exe_cls(**exe_kwargs) as executor:
        graphql(
            starwars_schema,
            HERO_QUERY,
            middlewares=[path_collector_one_way],
            executor=executor,
        )

    assert log == [
        "> hero",
        "> hero.id",
        "> hero.name",
        "> hero.friends",
        "> hero.friends[0].name",
        "> hero.friends[1].name",
        "> hero.friends[2].name",
    ]


def test_generator_middleware(starwars_schema):
    log = []

    def path_collector(next_, root, args, context, info):
        log.append("> %s" % stringify_path(info.path))
        yield next_(root, args, context, info)
        log.append("< %s" % stringify_path(info.path))

    graphql(starwars_schema, HERO_QUERY, middlewares=[path_collector])

    assert log == [
        "> hero",
        "> hero.id",
        "< hero.id",
        "> hero.name",
        "< hero.name",
        "> hero.friends",
        "> hero.friends[0].name",
        "< hero.friends[0].name",
        "> hero.friends[1].name",
        "< hero.friends[1].name",
        "> hero.friends[2].name",
        "< hero.friends[2].name",
        "< hero.friends",
        "< hero",
    ]


def test_generator_middleware_with_threads(starwars_schema):
    log = set()

    def path_collector(next_, root, args, context, info):
        log.add("> %s" % stringify_path(info.path))
        yield next_(root, args, context, info)
        log.add("< %s" % stringify_path(info.path))

    with ThreadPoolExecutor(max_workers=10) as executor:
        graphql(
            starwars_schema,
            HERO_QUERY,
            middlewares=[path_collector],
            executor=executor,
        )

    assert log == set(
        [
            "> hero",
            "> hero.id",
            "< hero.id",
            "> hero.name",
            "< hero.name",
            "> hero.friends",
            "> hero.friends[0].name",
            "< hero.friends[0].name",
            "> hero.friends[1].name",
            "< hero.friends[1].name",
            "> hero.friends[2].name",
            "< hero.friends[2].name",
            "< hero.friends",
            "< hero",
        ]
    )


class PathCollectorMiddleware(object):
    def __init__(self):
        self.log = []

    def __call__(self, next_, root, args, context, info):
        self.log.append("> %s" % stringify_path(info.path))
        yield next_(root, args, context, info)
        self.log.append("< %s" % stringify_path(info.path))


def test_class_based_middleware(starwars_schema):
    path_collector = PathCollectorMiddleware()
    graphql(starwars_schema, HERO_QUERY, middlewares=[path_collector])

    assert path_collector.log == [
        "> hero",
        "> hero.id",
        "< hero.id",
        "> hero.name",
        "< hero.name",
        "> hero.friends",
        "> hero.friends[0].name",
        "< hero.friends[0].name",
        "> hero.friends[1].name",
        "< hero.friends[1].name",
        "> hero.friends[2].name",
        "< hero.friends[2].name",
        "< hero.friends",
        "< hero",
    ]



# -*- coding: utf-8 -*-
"""
Mixed bag of exposed utility functions and classes that are used internally and
can be useful if you are building custom GraphQL tooling on top of this
library.
"""

# flake8: noqa

from .ast_node_from_value import ast_node_from_value
from .coerce_value import (
    coerce_argument_values,
    coerce_value,
    coerce_variable_values,
    directive_arguments,
)
from .default_resolver import default_resolver
from .introspection_query import introspection_query
from .type_info import TypeInfoVisitor
from .untyped_value_from_ast import untyped_value_from_ast
from .value_from_ast import value_from_ast

__all__ = (
    "ast_node_from_value",
    "coerce_argument_values",
    "coerce_value",
    "coerce_variable_values",
    "default_resolver",
    "directive_arguments",
    "introspection_query",
    "untyped_value_from_ast",
    "value_from_ast",
    "TypeInfoVisitor",
)



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

from py_gql.validation.rules import NoUndefinedVariablesChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_all_variables_defined(schema):
    run_test(
        NoUndefinedVariablesChecker,
        schema,
        """
        query Foo($a: String, $b: String, $c: String) {
            field(a: $a, b: $b, c: $c)
        }
        """,
    )


def test_all_variables_deeply_defined(schema):
    run_test(
        NoUndefinedVariablesChecker,
        schema,
        """
        query Foo($a: String, $b: String, $c: String) {
            field(a: $a) {
                field(b: $b) {
                    field(c: $c)
                }
            }
        }
        """,
    )


def test_all_variables_deeply_in_inline_fragments_defined(schema):
    run_test(
        NoUndefinedVariablesChecker,
        schema,
        """
        query Foo($a: String, $b: String, $c: String) {
            ... on Type {
                field(a: $a) {
                    field(b: $b) {
                        ... on Type {
                        field(c: $c)
                        }
                    }
                }
            }
        }
        """,
    )


def test_all_variables_in_fragments_deeply_defined(schema):
    run_test(
        NoUndefinedVariablesChecker,
        schema,
        """
        query Foo($a: String, $b: String, $c: String) {
            ...FragA
        }

        fragment FragA on Type {
            field(a: $a) {
                ...FragB
            }
        }

        fragment FragB on Type {
            field(b: $b) {
                ...FragC
            }
        }

        fragment FragC on Type {
            field(c: $c)
        }
        """,
    )


def test_variable_within_single_fragment_defined_in_multiple_operations(schema):
    run_test(
        NoUndefinedVariablesChecker,
        schema,
        """
        query Foo($a: String) {
            ...FragA
        }
        query Bar($a: String) {
            ...FragA
        }
        fragment FragA on Type {
            field(a: $a)
        }
        """,
    )


def test_variable_within_fragments_defined_in_operations(schema):
    run_test(
        NoUndefinedVariablesChecker,
        schema,
        """
        query Foo($a: String) {
            ...FragA
        }
        query Bar($b: String) {
            ...FragB
        }
        fragment FragA on Type {
            field(a: $a)
        }
        fragment FragB on Type {
            field(b: $b)
        }
        """,
    )


def test_variable_within_recursive_fragment_defined(schema):
    run_test(
        NoUndefinedVariablesChecker,
        schema,
        """
        query Foo($a: String) {
            ...FragA
        }

        fragment FragA on Type {
            field(a: $a) {
                ...FragA
            }
        }
        """,
        [],
    )


def test_variable_not_defined(schema):
    run_test(
        NoUndefinedVariablesChecker,
        schema,
        """
        query Foo($a: String, $b: String, $c: String) {
            field(a: $a, b: $b, c: $c, d: $d)
        }
        """,
        ['Variable "$d" is not defined on "Foo" operation'],
        [(99, 101)],
    )


def test_variable_not_defined_by_unnamed_query(schema):
    run_test(
        NoUndefinedVariablesChecker,
        schema,
        """
        {
            field(a: $a)
        }
        """,
        ['Variable "$a" is not defined on anonymous operation'],
        [(32, 34)],
    )


def test_multiple_variables_not_defined(schema):
    run_test(
        NoUndefinedVariablesChecker,
        schema,
        """
        query Foo($b: String) {
            field(a: $a, b: $b, c: $c)
        }
        """,
        [
            'Variable "$a" is not defined on "Foo" operation',
            'Variable "$c" is not defined on "Foo" operation',
        ],
    )


def test_variable_in_fragment_not_defined_by_unnamed_query(schema):
    run_test(
        NoUndefinedVariablesChecker,
        schema,
        """
        {
            ...FragA
        }
        fragment FragA on Type {
            field(a: $a)
        }
        """,
        [
            'Variable "$a" from fragment "FragA" is not defined on anonymous'
            " operation"
        ],
    )


def test_variable_in_fragment_not_defined_by_operation(schema):
    run_test(
        NoUndefinedVariablesChecker,
        schema,
        """
        query Foo($a: String, $b: String) {
            ...FragA
        }
        fragment FragA on Type {
            field(a: $a) {
                ...FragB
            }
        }
        fragment FragB on Type {
            field(b: $b) {
                ...FragC
            }
        }
        fragment FragC on Type {
            field(c: $c)
        }
        """,
        [
            'Variable "$c" from fragment "FragC" is not defined on "Foo" operation'
        ],
    )


def test_multiple_variables_in_fragments_not_defined(schema):
    run_test(
        NoUndefinedVariablesChecker,
        schema,
        """
        query Foo($b: String) {
            ...FragA
        }
        fragment FragA on Type {
            field(a: $a) {
                ...FragB
            }
        }
        fragment FragB on Type {
            field(b: $b) {
                ...FragC
            }
        }
        fragment FragC on Type {
            field(c: $c)
        }
        """,
        [
            'Variable "$a" from fragment "FragA" is not defined on "Foo" operation',
            'Variable "$c" from fragment "FragC" is not defined on "Foo" operation',
        ],
    )


def test_single_variable_in_fragment_not_defined_by_multiple_operations(schema):
    run_test(
        NoUndefinedVariablesChecker,
        schema,
        """
        query Foo($b: String) {
            ...FragAB
        }
        query Bar($a: String) {
            ...FragAB
        }
        fragment FragAB on Type {
            field(a: $a, b: $b)
        }
        """,
        [
            'Variable "$a" from fragment "FragAB" is not defined on "Foo" '
            "operation",
            'Variable "$b" from fragment "FragAB" is not defined on "Bar" '
            "operation",
        ],
    )


def test_variables_in_fragment_not_defined_by_multiple_operations(schema):
    run_test(
        NoUndefinedVariablesChecker,
        schema,
        """
        query Foo($b: String) {
            ...FragA
        }
        query Bar($a: String) {
            ...FragB
        }
        fragment FragA on Type {
            field(a: $a)
        }
        fragment FragB on Type {
            field(b: $b)
        }
        """,
        [
            'Variable "$a" from fragment "FragA" is not defined on "Foo" operation',
            'Variable "$b" from fragment "FragB" is not defined on "Bar" operation',
        ],
    )


def test_variable_in_fragment_used_by_other_operation(schema):
    run_test(
        NoUndefinedVariablesChecker,
        schema,
        """
        query Foo($b: String) {
            ...FragA
        }
        query Bar($a: String) {
            ...FragB
        }
        fragment FragA on Type {
            field(a: $a)
        }
        fragment FragB on Type {
            field(b: $b)
        }
        """,
        [
            'Variable "$a" from fragment "FragA" is not defined on "Foo" operation',
            'Variable "$b" from fragment "FragB" is not defined on "Bar" operation',
        ],
    )


def test_multiple_undefined_variables_produce_multiple_errors(schema):
    run_test(
        NoUndefinedVariablesChecker,
        schema,
        """
        query Foo($b: String) {
            ...FragAB
        }
        query Bar($a: String) {
            ...FragAB
        }
        fragment FragAB on Type {
            field1(a: $a, b: $b)
            ...FragC
            field3(a: $a, b: $b)
        }
        fragment FragC on Type {
            field2(c: $c)
        }
        """,
        [
            'Variable "$a" from fragment "FragAB" is not defined on "Foo" operation',
            'Variable "$c" from fragment "FragC" is not defined on "Foo" operation',
            'Variable "$b" from fragment "FragAB" is not defined on "Bar" operation',
            'Variable "$c" from fragment "FragC" is not defined on "Bar" operation',
        ],
    )



# -*- coding: utf-8 -*-

import pytest

from py_gql._string_utils import dedent
from py_gql.execution import GraphQLExtension, GraphQLResult


def test_GraphQLResult_response_with_nothing():
    assert GraphQLResult().response() == {}


def test_GraphQLResult_json():
    assert GraphQLResult(data={"foo": 42}).json(indent=4) == dedent(
        """
        {
            "data": {
                "foo": 42
            }
        }"""
    )


def test_GraphQLResult_bool():
    assert GraphQLResult(data={"foo": 42})
    assert not GraphQLResult(errors=["foo"])
    assert not GraphQLResult(data={"foo": None}, errors=["foo"])


def test_GraphQLResult_add_extension():
    class Ext(GraphQLExtension):
        def name(self):
            return "foo"

        def payload(self):
            return {"bar": "baz"}

    r = GraphQLResult()
    r.add_extension(Ext())
    assert r.response() == {"extensions": {"foo": {"bar": "baz"}}}


def test_GraphQLResult_add_extension_raises_on_duplicate():
    class Ext(GraphQLExtension):
        def name(self):
            return "foo"

        def payload(self):
            return {"bar": "baz"}

    r = GraphQLResult()
    r.add_extension(Ext())
    with pytest.raises(ValueError) as exc_info:
        r.add_extension(Ext())

    assert str(exc_info.value) == 'Duplicate extension "foo"'



# -*- coding: utf-8 -*-
""" Performance / Observability hooks for GraphQL execution.
"""

import contextlib


class GraphQLTracer(object):
    """ Subclass this to implement custom tracing.

    You can either override the :meth:`trace` method or the specific
    ``on_.*_(start|end)`` event handlers.
    """

    def trace(self, evt, stage, **kwargs):
        """ Register a trace event.

        The default behaviour is to dispatch to other methods.

        Args:
            evt (str): Event type
            stage (str): "start" or "end"
            **kwargs: Event payload
        """
        getattr(self, "on_%s_%s" % (evt, stage))(**kwargs)

    @contextlib.contextmanager
    def _trace_context(self, evt, **kwargs):
        self.trace(evt, "start", **kwargs)
        try:
            yield
        finally:
            self.trace(evt, "end", **kwargs)

    def _middleware(self, next_step, root, args, context, info):
        self.trace("field", "start", args=args, info=info)
        yield next_step(root, args, context, info)
        self.trace("field", "end", args=args, info=info)

    def on_query_start(self, document, variables, operation_name):
        """ Called before graphql processing starts.

        Args:
            document (str): The query document as a string
            variables (Optiona[dict]): Raw JSON Variables
            operation_name (Optiona[str]): Operation name
        """
        pass

    def on_query_end(self, document, variables, operation_name):
        """ Called after graphql processing ends.

        Args:
            document (str): The query document as a string
            variables (Optiona[dict]): Raw JSON Variables
            operation_name (Optiona[str]): Operation name
        """
        pass

    def on_parse_start(self, document):
        """ Called before graphql parsing starts.

        Args:
            document (str): The query document as a string
        """
        pass

    def on_parse_end(self, document):
        """ Called after graphql parsing ends.

        Args:
            document (str): The query document as a string
        """
        pass

    def on_validate_start(self, ast):
        """ Called before graphql validation starts.

        Args:
            ast (py_gql.lang.ast.Document): The parsed query
        """
        pass

    def on_validate_end(self, ast):
        """ Called after graphql validation ends.

        Args:
            ast (py_gql.lang.ast.Document): The parsed query
        """
        pass

    def on_execute_start(self, ast, variables):
        """ Called before graphql execution starts.

        Args:
            ast (py_gql.lang.ast.Document): The parsed query
            variables (Optiona[dict]): Raw JSON Variables
        """
        pass

    def on_execute_end(self, ast, variables):
        """ Called after graphql execution ends.

        Args:
            ast (py_gql.lang.ast.Document): The parsed query
            variables (Optiona[dict]): Raw JSON Variables
        """
        pass

    def on_field_start(self, args, info):
        """ Called before graphql field resolution starts.

        Args:
            args (dict): Coereced field arguments
            info (py_gql.execution.ResolveInfo): Resolution context
        """
        pass

    def on_field_end(self, args, info):
        """ Called after graphql field resolution is complete, including
        all child fields.

        Args:
            args (dict): Coereced field arguments
            info (py_gql.execution.ResolveInfo): Resolution context
        """
        pass


class NullTracer(GraphQLTracer):
    """ Default noop tracer
    """

    def trace(self, *args, **kwargs):
        pass



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import UniqueArgumentNamesChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_no_arguments_on_field(schema):
    run_test(
        UniqueArgumentNamesChecker,
        schema,
        """
    {
        field
    }
    """,
        [],
        [],
    )


def test_no_arguments_on_directive(schema):
    run_test(
        UniqueArgumentNamesChecker,
        schema,
        """
    {
        field @directive
    }
    """,
        [],
        [],
    )


def test_argument_on_field(schema):
    run_test(
        UniqueArgumentNamesChecker,
        schema,
        """
    {
        field(arg: "value")
    }
    """,
        [],
        [],
    )


def test_argument_on_directive(schema):
    run_test(
        UniqueArgumentNamesChecker,
        schema,
        """
    {
        field @directive(arg: "value")
    }
    """,
        [],
        [],
    )


def test_same_argument_on_two_fields(schema):
    run_test(
        UniqueArgumentNamesChecker,
        schema,
        """
    {
        one: field(arg: "value")
        two: field(arg: "value")
    }
    """,
        [],
        [],
    )


def test_same_argument_on_field_and_directive(schema):
    run_test(
        UniqueArgumentNamesChecker,
        schema,
        """
    {
        field(arg: "value") @directive(arg: "value")
    }
    """,
        [],
        [],
    )


def test_same_argument_on_two_directives(schema):
    run_test(
        UniqueArgumentNamesChecker,
        schema,
        """
    {
        field @directive1(arg: "value") @directive2(arg: "value")
    }
    """,
        [],
        [],
    )


def test_multiple_field_arguments(schema):
    run_test(
        UniqueArgumentNamesChecker,
        schema,
        """
    {
        field(arg1: "value", arg2: "value", arg3: "value")
    }
    """,
        [],
        [],
    )


def test_multiple_directive_arguments(schema):
    run_test(
        UniqueArgumentNamesChecker,
        schema,
        """
    {
        field @directive(arg1: "value", arg2: "value", arg3: "value")
    }
    """,
        [],
        [],
    )


def test_duplicate_field_arguments(schema):
    run_test(
        UniqueArgumentNamesChecker,
        schema,
        """
    {
        field(arg1: "value", arg1: "value")
    }
    """,
        ['Duplicate argument "arg1"'],
        [(36, 49)],
    )


def test_many_duplicate_field_arguments(schema):
    run_test(
        UniqueArgumentNamesChecker,
        schema,
        """
    {
        field(arg1: "value", arg1: "value", arg1: "value")
    }
    """,
        ['Duplicate argument "arg1"', 'Duplicate argument "arg1"'],
        [(36, 49), (51, 64)],
    )


def test_duplicate_directive_arguments(schema):
    run_test(
        UniqueArgumentNamesChecker,
        schema,
        """
    {
        field @directive(arg1: "value", arg1: "value")
    }
    """,
        ['Duplicate argument "arg1"'],
    )


def test_many_duplicate_directive_arguments(schema):
    run_test(
        UniqueArgumentNamesChecker,
        schema,
        """
    {
        field @directive(arg1: "value", arg1: "value", arg1: "value")
    }
    """,
        ['Duplicate argument "arg1"', 'Duplicate argument "arg1"'],
    )



# -*- coding: utf-8 -*-
""" """

# flake8: noqa

from ._execute import execute
from .tracing import GraphQLTracer
from .wrappers import GraphQLExtension, GraphQLResult, ResolveInfo

__all__ = (
    "GraphQLExtension",
    "GraphQLResult",
    "GraphQLTracer",
    "ResolveInfo",
    "execute",
)



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import UniqueVariableNamesChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_unique_variable_names(schema):
    run_test(
        UniqueVariableNamesChecker,
        schema,
        """
        query A($x: Int, $y: String) { __typename }
        query B($x: String, $y: Int) { __typename }
    """,
    )


def test_duplicate_variable_names(schema):
    run_test(
        UniqueVariableNamesChecker,
        schema,
        """
        query A($x: Int, $x: Int, $x: String) { __typename }
        query B($x: String, $x: Int) { __typename }
        query C($x: Int, $x: Int) { __typename }
    """,
        [
            'Duplicate variable "$x"',
            'Duplicate variable "$x"',
            'Duplicate variable "$x"',
            'Duplicate variable "$x"',
        ],
        [(26, 33), (35, 45), (90, 97), (139, 146)],
    )



# -*- coding: utf-8 -*-

import pytest

from py_gql.schema import (
    ID,
    Argument,
    Boolean,
    Directive,
    EnumType,
    EnumValue,
    Field,
    Float,
    InputField,
    InputObjectType,
    Int,
    InterfaceType,
    ListType,
    NonNullType,
    ObjectType,
    ScalarType,
    Schema,
    String,
    UnionType,
)


@pytest.fixture
def schema():

    Being = InterfaceType(
        "Being", [Field("name", String, [Argument("surname", Boolean)])]
    )

    Pet = InterfaceType(
        "Pet", [Field("name", String, [Argument("surname", Boolean)])]
    )

    Canine = InterfaceType(
        "Canine", [Field("name", String, [Argument("surname", Boolean)])]
    )

    DogCommand = EnumType(
        "DogCommand",
        [EnumValue("SIT", 0), EnumValue("HEEL", 1), EnumValue("DOWN", 2)],
    )

    FurColor = EnumType(
        "FurColor",
        [
            EnumValue("BROWN", 0),
            EnumValue("BLACK", 1),
            EnumValue("TAN", 2),
            EnumValue("SPOTTED", 3),
            EnumValue("NO_FUR", None),
            EnumValue("UNKNOWN", -1),
        ],
    )

    Dog = ObjectType(
        "Dog",
        [
            Field("name", String, args=[Argument("surname", Boolean)]),
            Field("nickname", String),
            Field("barkVolume", Int),
            Field("barks", Boolean),
            Field(
                "doesKnowCommand", Boolean, [Argument("dogCommand", DogCommand)]
            ),
            Field(
                "isHousetrained",
                Boolean,
                [Argument("atOtherHomes", Boolean, True)],
            ),
            Field(
                "isAtLocation",
                Boolean,
                [Argument("x", Int), Argument("y", Int)],
            ),
        ],
        [Being, Pet, Canine],
    )

    Cat = ObjectType(
        "Cat",
        [
            Field("name", String, args=[Argument("surname", Boolean)]),
            Field("nickname", String),
            Field("meowVolume", Int),
            Field("meows", Boolean),
            Field("furColor", FurColor),
        ],
        [Being, Pet],
    )

    CatOrDog = UnionType("CatOrDog", [Dog, Cat])

    Intelligent = InterfaceType("Intelligent", [Field("iq", Int)])

    Human = ObjectType(
        "Human",
        lambda: [
            Field("name", String, args=[Argument("surname", Boolean)]),
            Field("iq", Int),
            Field("pets", ListType(Pet)),
            Field("relatives", ListType(Human)),
        ],
        [Being, Intelligent],
    )

    Alien = ObjectType(
        "Alien",
        [
            Field("name", String, args=[Argument("surname", Boolean)]),
            Field("iq", Int),
            Field("numEyes", Int),
        ],
        [Being, Intelligent],
    )

    DogOrHuman = UnionType("DogOrHuman", [Dog, Human])

    HumanOrAlien = UnionType("HumanOrAlien", [Human, Alien])

    ComplexInput = InputObjectType(
        "ComplexInput",
        [
            InputField("requiredField", NonNullType(Boolean)),
            InputField(
                "nonNullField", NonNullType(Boolean), default_value=False
            ),
            InputField("intField", Int),
            InputField("stringField", String),
            InputField("booleanField", Boolean),
            InputField("stringListField", ListType(String)),
        ],
    )

    ComplicatedArgs = ObjectType(
        "ComplicatedArgs",
        [
            Field("intArgField", String, [Argument("intArg", Int)]),
            Field(
                "nonNullIntArgField",
                String,
                [Argument("nonNullIntArg", NonNullType(Int))],
            ),
            Field("stringArgField", String, [Argument("stringArg", String)]),
            Field("booleanArgField", String, [Argument("booleanArg", Boolean)]),
            Field("enumArgField", String, [Argument("enumArg", FurColor)]),
            Field("floatArgField", String, [Argument("floatArg", Float)]),
            Field("idArgField", String, [Argument("idArg", ID)]),
            Field(
                "stringListArgField",
                String,
                [Argument("stringListArg", ListType(String))],
            ),
            Field(
                "stringListNonNullArgField",
                String,
                [
                    Argument(
                        "stringListNonNullArg", ListType(NonNullType(String))
                    )
                ],
            ),
            Field(
                "complexArgField",
                String,
                [Argument("complexArg", ComplexInput)],
            ),
            Field(
                "multipleReqs",
                String,
                [
                    Argument("req1", NonNullType(Int)),
                    Argument("req2", NonNullType(Int)),
                ],
            ),
            Field(
                "nonNullFieldWithDefault",
                String,
                [Argument("arg", NonNullType(Int), default_value=0)],
            ),
            Field(
                "multipleOpts",
                String,
                [Argument("opt1", Int, 0), Argument("opt2", Int, 0)],
            ),
            Field(
                "multipleOptAndReq",
                String,
                [
                    Argument("req1", NonNullType(Int)),
                    Argument("req2", NonNullType(Int)),
                    Argument("opt1", Int, 0),
                    Argument("opt2", Int, 0),
                ],
            ),
        ],
    )

    def _invalid(*args, **kwargs):
        raise ValueError("Invalid scalar is always invalid")

    InvalidScalar = ScalarType("Invalid", lambda x: x, _invalid, _invalid)
    AnyScalar = ScalarType("Any", lambda x: x, lambda x: x, lambda x: x)

    return Schema(
        ObjectType(
            "QueryRoot",
            [
                Field("human", Human, [Argument("id", ID)]),
                Field("alien", Alien),
                Field("dog", Dog),
                Field("cat", Cat),
                Field("pet", Pet),
                Field("catOrDog", CatOrDog),
                Field("dogOrHuman", DogOrHuman),
                Field("humanOrAlien", HumanOrAlien),
                Field("humanOrAlien", HumanOrAlien),
                Field("complicatedArgs", ComplicatedArgs),
                Field("invalidArg", String, [Argument("arg", InvalidScalar)]),
                Field("anydArg", String, [Argument("arg", AnyScalar)]),
            ],
        ),
        types=[Cat, Dog, Human, Alien],
        directives=[
            Directive("onQuery", ["QUERY"]),
            Directive("onMutation", ["MUTATION"]),
            Directive("onSubscription", ["SUBSCRIPTION"]),
            Directive("onField", ["FIELD"]),
            Directive("onFragmentDefinition", ["FRAGMENT_DEFINITION"]),
            Directive("onFragmentSpread", ["FRAGMENT_SPREAD"]),
            Directive("onInlineFragment", ["INLINE_FRAGMENT"]),
            Directive("onSchema", ["SCHEMA"]),
            Directive("onScalar", ["SCALAR"]),
            Directive("onObject", ["OBJECT"]),
            Directive("onFieldDefinition", ["FIELD_DEFINITION"]),
            Directive("onArgumentDefinition", ["ARGUMENT_DEFINITION"]),
            Directive("onInterface", ["INTERFACE"]),
            Directive("onUnion", ["UNION"]),
            Directive("onEnum", ["ENUM"]),
            Directive("onEnumValue", ["ENUM_VALUE"]),
            Directive("onInputObject", ["INPUT_OBJECT"]),
            Directive("onInputFieldDefinition", ["INPUT_FIELD_DEFINITION"]),
        ],
    )


@pytest.fixture
def schema_2():
    SomeBox = InterfaceType(
        "SomeBox",
        [Field("deepBox", lambda: SomeBox), Field("unrelatedField", String)],
    )

    StringBox = ObjectType(
        "StringBox",
        [
            Field("scalar", String),
            Field("deepBox", lambda: StringBox),
            Field("unrelatedField", String),
            Field("listStringBox", lambda: ListType(StringBox)),
            Field("stringBox", lambda: StringBox),
            Field("intBox", lambda: IntBox),
        ],
        [SomeBox],
    )

    IntBox = ObjectType(
        "IntBox",
        [
            Field("scalar", Int),
            Field("deepBox", lambda: IntBox),
            Field("unrelatedField", String),
            Field("listStringBox", lambda: ListType(StringBox)),
            Field("stringBox", lambda: StringBox),
            Field("intBox", lambda: IntBox),
        ],
        [SomeBox],
    )

    NonNullStringBox1 = InterfaceType(
        "NonNullStringBox1", [Field("scalar", NonNullType(String))]
    )

    NonNullStringBox1Impl = ObjectType(
        "NonNullStringBox1Impl",
        [
            Field("scalar", NonNullType(String)),
            Field("unrelatedField", String),
            Field("deepBox", SomeBox),
        ],
        [SomeBox, NonNullStringBox1],
    )

    NonNullStringBox2 = InterfaceType(
        "NonNullStringBox2", [Field("scalar", NonNullType(String))]
    )

    NonNullStringBox2Impl = ObjectType(
        "NonNullStringBox2Impl",
        [
            Field("scalar", NonNullType(String)),
            Field("unrelatedField", String),
            Field("deepBox", SomeBox),
        ],
        [SomeBox, NonNullStringBox2],
    )

    Connection = ObjectType(
        "Connection",
        [
            Field(
                "edges",
                ListType(
                    ObjectType(
                        "Edge",
                        [
                            Field(
                                "node",
                                ObjectType(
                                    "Node",
                                    [Field("id", ID), Field("name", String)],
                                ),
                            )
                        ],
                    )
                ),
            )
        ],
    )

    yield Schema(
        ObjectType(
            "QueryRoot",
            [Field("someBox", SomeBox), Field("connection", Connection)],
        ),
        types=[IntBox, StringBox, NonNullStringBox1Impl, NonNullStringBox2Impl],
    )



# -*- coding: utf-8 -*-

import pytest

from py_gql.exc import CoercionError
from py_gql.lang import ast as _ast
from py_gql.schema import Argument, Field, Int, NonNullType
from py_gql.utilities import coerce_argument_values


def _test_node(argument_value=None):
    if argument_value is None:
        arguments = []
    else:
        arguments = [
            _ast.Argument(name=_ast.Name(value="foo"), value=argument_value)
        ]
    return _ast.Field(name=_ast.Name(value="test"), arguments=arguments)


def _var(name):
    return _ast.Variable(name=_ast.Name(value=name))


def test_missing_nullable_arg_with_default():
    arg = Argument("foo", Int, default_value=42)
    field = Field("test", Int, [arg])
    node = _test_node()
    assert coerce_argument_values(field, node) == {"foo": 42}


def test_missing_nullable_arg_without_default():
    arg = Argument("foo", Int)
    field = Field("test", Int, [arg])
    node = _test_node()
    assert coerce_argument_values(field, node) == {}


def test_missing_non_nullable_arg_with_default():
    arg = Argument("foo", NonNullType(Int), default_value=42)
    field = Field("test", Int, [arg])
    node = _test_node()
    assert coerce_argument_values(field, node) == {"foo": 42}


def test_missing_non_nullable_arg_without_default():
    arg = Argument("foo", NonNullType(Int))
    field = Field("test", Int, [arg])
    node = _test_node()
    with pytest.raises(CoercionError) as exc_info:
        coerce_argument_values(field, node)
    assert (
        str(exc_info.value)
        == 'Argument "foo" of required type "Int!" was not provided'
    )


def test_provided_value():
    arg = Argument("foo", Int)
    field = Field("test", Int, [arg])
    node = _test_node(_ast.IntValue(value="42"))
    assert coerce_argument_values(field, node) == {"foo": 42}


def test_provided_invalid_value():
    arg = Argument("foo", Int)
    field = Field("test", Int, [arg])
    node = _test_node(_ast.StringValue(value="foo"))
    with pytest.raises(CoercionError) as exc_info:
        assert coerce_argument_values(field, node)
    assert str(exc_info.value) == (
        'Argument "foo" of type "Int" was provided invalid value "foo" '
        "(Invalid literal StringValue)"
    )


def test_provided_known_variable():
    arg = Argument("foo", Int)
    field = Field("test", Int, [arg])
    node = _test_node(_var("bar"))
    assert coerce_argument_values(field, node, {"bar": 42}) == {"foo": 42}


def test_provided_unknown_variable_without_default_nullable():
    arg = Argument("foo", Int)
    field = Field("test", Int, [arg])
    node = _test_node(_var("bar"))
    assert coerce_argument_values(field, node) == {}


def test_provided_unknown_variable_without_default_non_nullable():
    arg = Argument("foo", NonNullType(Int))
    field = Field("test", Int, [arg])
    node = _test_node(_var("bar"))
    with pytest.raises(CoercionError) as exc_info:
        coerce_argument_values(field, node)
    assert str(exc_info.value) == (
        'Argument "foo" of required type "Int!" was provided the missing '
        'variable "$bar"'
    )



# -*- coding: utf-8 -*-
""" """

import pytest

from py_gql.exc import CoercionError
from py_gql.schema import (
    EnumType,
    Float,
    InputField,
    InputObjectType,
    Int,
    ListType,
    NonNullType,
    String,
)
from py_gql.utilities import coerce_value


def _test(value, type_, expected_result, expected_error=None):
    if expected_error is not None:
        with pytest.raises(CoercionError) as exc_info:
            coerce_value(value, type_)
        assert str(exc_info.value) == expected_error
    else:
        assert coerce_value(value, type_) == expected_result


def test_String_raises_on_list():
    _test(
        [1, 2, 3],
        String,
        None,
        'String cannot represent list value "[1, 2, 3]"',
    )


def test_Int_from_int_input():
    _test("1", Int, 1)


def test_Int_from_int_input_1():
    _test(1, Int, 1)


def test_Int_from_negative_int_input():
    _test("-1", Int, -1)


def test_Int_from_exponent_input():
    _test("1e3", Int, 1e3)


def test_Int_from_null_value():
    _test(None, Int, None)


def test_Int_raises_for_empty_value():
    _test(
        "",
        Int,
        None,
        "Int cannot represent non 32-bit signed integer: (empty string)",
    )


def test_Int_raises_for_float_input():
    _test("1.5", Int, None, "Int cannot represent non-integer value: 1.5")


def test_Int_raises_for_char_input():
    _test("c", Int, None, "Int cannot represent non-integer value: c")


def test_Int_raises_for_string_input():
    _test("meow", Int, None, "Int cannot represent non-integer value: meow")


def test_Float_for_int_input():
    _test("1", Float, 1.0)


def test_Float_for_exponent_input():
    _test("1e3", Float, 1e3)


def test_Float_for_float_input():
    _test("1.5", Float, 1.5)


def test_Float_raises_for_empty_value():
    _test(
        "",
        Float,
        None,
        "Float cannot represent non numeric value: (empty string)",
    )


def test_Float_raises_for_char_input():
    _test("a", Float, None, "Float cannot represent non numeric value: a")


def test_Float_raises_for_string_input():
    _test("meow", Float, None, "Float cannot represent non numeric value: meow")


Enum = EnumType("TestEnum", [("FOO", "InternalFoo"), ("BAR", 123456789)])


def test_Enum_for_a_known_enum_names():
    _test("FOO", Enum, "InternalFoo")
    _test("BAR", Enum, 123456789)


def test_Enum_raises_for_misspelled_enum_value():
    _test("foo", Enum, None, "Invalid name foo for enum TestEnum")


def test_Enum_raises_for_incorrect_value_type():
    _test(123, Enum, None, "Expected type TestEnum")


Input = InputObjectType(
    "TestInputObject",
    [InputField("foo", NonNullType(Int)), InputField("bar", Int)],
)


def test_InputObject_for_valid_input():
    _test({"foo": 123}, Input, {"foo": 123})


def test_InputObject_raises_for_non_dict_input():
    _test(123, Input, None, "Expected type TestInputObject to be an object")


def test_InputObject_raises_for_invalid_field():
    _test(
        {"foo": "abc"},
        Input,
        None,
        "Int cannot represent non-integer value: abc at value.foo",
    )


def test_InputObject_raises_for_missing_required_field():
    _test(
        {},
        Input,
        None,
        "Field foo of required type Int! was not provided at value.foo",
    )


def test_InputObject_raises_for_unknown_field():
    _test(
        {"foo": 1, "baz": 1},
        Input,
        None,
        "Field baz is not defined by type TestInputObject",
    )


def test_ListType_for_single_valid_value():
    _test("1", ListType(Int), [1])


def test_ListType_for_valid_values():
    _test([1, "2", "3"], ListType(Int), [1, 2, 3])


def test_ListType_raises_for_invalid_item():
    _test(
        [1, "abc", "3"],
        ListType(Int),
        None,
        "Int cannot represent non-integer value: abc at value[1]",
    )


def test_ListType_raises_for_invalid_items():
    _test(
        [1, "abc", "def"],
        ListType(Int),
        None,
        "Int cannot represent non-integer value: abc at value[1],\n"
        "Int cannot represent non-integer value: def at value[2]",
    )


def test_nested_error():
    _test(
        [{"foo": "abc"}],
        ListType(Input),
        None,
        "Int cannot represent non-integer value: abc at value[0].foo",
    )


def test_null_non_nullable_type():
    _test(
        None,
        NonNullType(Int),
        None,
        "Expected non-nullable type Int! not to be null",
    )



# -*- coding: utf-8 -*-
""" Test default validation. """

from ._test_utils import assert_validation_result


def test_it_validates_queries(schema):
    assert_validation_result(
        schema,
        """
    query {
        catOrDog {
            ... on Cat {
                furColor
            }
            ... on Dog {
               isHousetrained
            }
        }
    }
    """,
    )


def test_it_detects_bad_scalar_parse(schema):
    assert_validation_result(
        schema,
        """
    query {
        invalidArg(arg: "bad value")
    }
    """,
        [
            'Expected type Invalid, found "bad value" '
            "(Invalid scalar is always invalid)"
        ],
    )


# Star Wars schema related tests


def test_complex_but_valid_query(starwars_schema):
    assert_validation_result(
        starwars_schema,
        """
    query NestedQueryWithFragment {
        hero {
            ...NameAndAppearances
            friends {
                ...NameAndAppearances
                friends {
                ...NameAndAppearances
                }
            }
        }
    }

    fragment NameAndAppearances on Character {
        name
        appearsIn
    }
    """,
    )


def test_non_existent_fields_are_invalid(starwars_schema):
    assert_validation_result(
        starwars_schema,
        """
    query HeroSpaceshipQuery {
        hero {
           favoriteSpaceship
        }
    }
    """,
        ['Cannot query field "favoriteSpaceship" on type "Character"'],
    )


def test_requires_fields_on_objects(starwars_schema):
    assert_validation_result(
        starwars_schema,
        """
    query HeroNoFieldsQuery {
        hero
    }
    """,
        ['Field "hero" of type "Character" must have a subselection'],
    )


def test_disallows_fields_on_scalars(starwars_schema):
    assert_validation_result(
        starwars_schema,
        """
    query HeroFieldsOnScalarQuery {
        hero {
            name {
                firstCharacterOfName
            }
        }
    }
    """,
        ['Field "name" cannot have a selection as type "String" has no fields'],
    )


def test_disallows_object_fields_on_interface(starwars_schema):
    assert_validation_result(
        starwars_schema,
        """
    query DroidFieldOnCharacter {
        hero {
            name
            primaryFunction
        }
    }
    """,
        ['Cannot query field "primaryFunction" on type "Character"'],
    )


def test_allows_object_fields_in_fragments(starwars_schema):
    assert_validation_result(
        starwars_schema,
        """
    query DroidFieldInFragment {
        hero {
            name
           ...DroidFields
        }
    }

    fragment DroidFields on Droid {
        primaryFunction
    }
    """,
    )


def test_allows_object_fields_in_inline_fragments(starwars_schema):
    assert_validation_result(
        starwars_schema,
        """
    query DroidFieldInFragment {
        hero {
            name
            ... on Droid {
                primaryFunction
            }
        }
    }
    """,
    )
